//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2005 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////////

//	This version has been modified by fccHandler, August 2007

#include <stddef.h>	// Meia
#include <stdio.h>
#include <process.h>
#include <crtdbg.h>
#include <assert.h>
#ifdef _WIN32
	#include <windows.h>
#endif
#include <math.h>
#include <vd2/system/atomic.h>	// Meia
#include <vd2/system/memory.h>	// Meia
#include <vd2/system/error.h>
#include <vd2/system/cpuaccel.h>

#include <vd2/Meia2/MPEG2Decoder.h>
#include <vd2/Meia2/MPEGConvert.h>
//#include <vd2/Meia2/MPEGIDCT.h>

#include "mpeg2_tables.h"
#include "mpeg2_asm.h"

// Support ATI's VCR format?
//#define ATIVCR_SUPPORT
// (doesn't work)

// Start codes
#define SEQUENCE_START_CODE		0x1B3
#define PICTURE_START_CODE		0x100
#define SLICE_START_CODE_MIN	0x101
#define SLICE_START_CODE_MAX	0x1AF
#define EXTENSION_START_CODE	0x1B5
#define USER_DATA_START_CODE	0x1B2

// Extension start code IDs
#define SEQUENCE_EXTENSION_ID			1
#define SEQUENCE_DISPLAY_EXTENSION_ID	2
#define QUANT_MATRIX_EXTENSION_ID		3
#define COPYRIGHT_EXTENSION_ID			4
#define PICTURE_DISPLAY_EXTENSION_ID	7
#define PICTURE_CODING_EXTENSION_ID		8

#define MPEG_FRAME_TYPE_I			(1)
#define MPEG_FRAME_TYPE_P			(2)
#define MPEG_FRAME_TYPE_B			(3)
#define MPEG_FRAME_TYPE_D			(4)

// Chroma formats
#define CHROMA420		1
#define CHROMA422		2
#define CHROMA444		3
// Picture structures
#define TOP_FIELD		1
#define BOTTOM_FIELD	2
#define FRAME_PICTURE	3
// Motion_types
#define MC_FIELD	1
#define MC_FRAME	2
#define MC_16X8		2
#define MC_DMV		3
// Motion vector types
#define MV_FIELD	0
#define MV_FRAME	1
// flags for MPEGBuffer.bitflags
#define PICTUREFLAG_RFF	1	// repeat_first_field
#define PICTUREFLAG_TFF	2	// top_field_first
#define PICTUREFLAG_PF	4	// progressive_frame

// Number of buffers we need
#define MPEG_BUFFERS	(4UL)

typedef unsigned char YCCSample;

typedef struct {
	YCCSample *Y;
	YCCSample *U;
	YCCSample *V;
	int frame_num;
	int bitflags;
} MPEGBuffer;


class VDMPEGDecoder : public IVDMPEGDecoder {
private:
/*
	typedef unsigned char YCCSample;

	enum {
		kMBF_NewQuant	= 16,
		kMBF_Both		= 12,
		kMBF_Forward	= 8,
		kMBF_Backward	= 4,
		kMBF_Pattern	= 2,
		kMBF_Intra		= 1
	};
*/
	VDAtomicInt mRefCount;

	LPVOID memblock;
	const unsigned char *bitarray;
	unsigned long totbits;
	unsigned long bitpos;

	unsigned char intramatrix[64];
	unsigned char nonintramatrix[64];
	unsigned char chromaintramatrix[64];
	unsigned char chromanonintramatrix[64];

	int forw_vector_full_pel;
	int back_vector_full_pel;

	bool MPEG2_Flag;
	int quantizer_scale;
	int matrix_coefficients;
	int chroma_format;
	int block_count;
	int progressive_sequence;
	int f_code[2][2];
	int intra_dc_precision;
	int picture_structure;
	int frame_pred_frame_dct;
	int concealment_vectors;
	int q_scale_type;
	int intra_vlc_format;
	int alternate_scan;
	long PMV[2][2][2];
	long dmvector[2];
	long mvfield_select[2][2];
	int motion_vector_count;
	int motion_type;
	int mv_format;
	int mvscale;
	int dct_type;

	bool fAllowMatrixCoefficients;
/*
	struct MPEGBuffer {
		YCCSample *Y;
		YCCSample *U;
		YCCSample *V;
		int frame_num;
		int bitflags;
	} buffers[MPEG_BUFFERS];
*/
	MPEGBuffer buffers[MPEG_BUFFERS];

	MPEGBuffer *current_frame;
	MPEGBuffer *forw_frame;
	MPEGBuffer *back_frame;
	MPEGBuffer *aux_frame;

	YCCSample *uv422, *u444, *v444;

	short (*dct_coeff)[64];

//	const VDMPEGIDCTSet	*mpIDCTs;

	int second_field;
	long mErrorState;
	long cpuflags;

	int frame_type;
	long pelWidth, pelHeight, mbWidth, mbHeight;
	long y_pitch;

	int dct_dc_y_past, dct_dc_u_past, dct_dc_v_past;
	int macro_block_flags;

	bool mpeg_alloc();
	long show_bits(int nBits);
	long get_bits(int nBits);
	int next_start_code();
	inline void picture_coding_extension();
	inline void quant_matrix_extension();
	void extension_and_user_data();
	inline bool sequence_header();
	inline void picture_header();
	inline void idctcol(short *blk);
	inline void idctrow(short *blk);
	void Fast_IDCT(void *src);
	void Add_Block(int comp, int bx, int by, int addflag);
	long get_luma_diff();
	long get_chroma_diff();
	inline void Decode_MPEG1_Intra_Block(int comp);
	inline void Decode_MPEG1_Non_Intra_Block(int comp);
	inline void Decode_MPEG2_Intra_Block(int comp);
	inline void Decode_MPEG2_Non_Intra_Block(int comp);

	void form_component_prediction(YCCSample* src, YCCSample* dst,
		int lx, int w, int h, int x, int y, int dx, int dy, int average_flag);

	void form_prediction(MPEGBuffer *srcBuf, int sfield, MPEGBuffer *dstBuf, int dfield,
		int lx, int w, int h, int x, int y, int dx, int dy, int average_flag);

	void dual_prime(int DMV[][2], int mvx, int mvy);
	void form_predictions(int bx, int by);
	int mpeg_get_motion_component();
	int get_mb_address_inc();
	inline int get_I_mb_flags();
	inline int get_P_mb_flags();
	inline int get_B_mb_flags();
	inline int get_D_mb_flags();
	inline int get_mb_flags();
	inline int get_mb_cbp();
	void decode_motion_vector(long& pred, int r_size, long code, long residual, int full_pel);
	void motion_vector(long *pmv, int h_size, int v_size, int full_pel);
	void motion_vectors(int s, int h_size, int v_size);
	inline void decode_mb(int MBA);
	inline void skipped_mb(int MBA);
	inline void slice(int pos);

	void mpeg_conv420to422(YCCSample *src, YCCSample *dst, long w, long h, bool progressive);
	void mpeg_conv422to444(YCCSample *src, YCCSample *dst, long w, long h, int mpeg2);

	inline void mpeg_conv444toRGB(YCCSample *srcY, YCCSample *srcU,
		YCCSample *srcV, void *_dst, int bpr, int w, int h, int bpp);
	
	void mpeg_YUVToRGB(int buffer_ID, void *dst, long bpr, long w, long h, int bpp);
	bool mpeg_YUVToYUY2(void *dst, ptrdiff_t pitch, int buffer,	YCCSample *Y_ptr, YCCSample *U_ptr, YCCSample *V_ptr);

public:
	VDMPEGDecoder();
	~VDMPEGDecoder() throw();
	
	int AddRef() { return mRefCount.inc(); }
	int Release() { int rc = mRefCount.dec(); if (!rc) delete this; return rc; }
	
	// decoding
	
	bool Init(int width, int height, long seq_ext);
	void Shutdown();
	void SetIntraQuantizers(const unsigned char *pMatrix);
	void SetNonintraQuantizers(const unsigned char *pMatrix);
	void SetPredictors(const VDMPEGPredictorSet *pPredictors);
	void SetConverters(const VDMPEGConverterSet *pConverters);
	void SetIDCTs(const VDMPEGIDCTSet *pIDCTs);
	void EnableMatrixCoefficients(bool bEnable);
	
	int DecodeFrame(const void *src, long len, long frame, int dst, int fwd, int rev);
	long GetErrorState();
	
	// framebuffer access
	
	int GetFrameBuffer(long frame);
	long GetFrameNumber(int buffer);
	void CopyFrameBuffer(int dst, int src, long frameno);
	void SwapFrameBuffers(int dst, int src);
	void ClearFrameBuffers();

	const void *GetYBuffer(int buffer, ptrdiff_t& pitch);
	const void *GetCrBuffer(int buffer, ptrdiff_t& pitch);
	const void *GetCbBuffer(int buffer, ptrdiff_t& pitch);

	void CopyField(int dst, int dstfield, int src, int srcfield);

	bool Decode(void *dst, long pitch, int buffer, tVDMPEGConverter pConv);
	bool DecodeUYVY(void *dst, ptrdiff_t pitch, int buffer);
	bool DecodeYUYV(void *dst, ptrdiff_t pitch, int buffer);
	bool DecodeYVYU(void *dst, ptrdiff_t pitch, int buffer);
	bool DecodeY41P(void *dst, ptrdiff_t pitch, int buffer);
	bool DecodeRGB15(void *dst, ptrdiff_t pitch, int buffer);
	bool DecodeRGB16(void *dst, ptrdiff_t pitch, int buffer);
	bool DecodeRGB24(void *dst, ptrdiff_t pitch, int buffer);
	bool DecodeRGB32(void *dst, ptrdiff_t pitch, int buffer);
};

#define SetError(x) mErrorState |= x

VDMPEGDecoder::VDMPEGDecoder()
	: mRefCount(0)
{
	memblock = NULL;
	fAllowMatrixCoefficients = true;
	cpuflags = CPUGetEnabledExtensions();
}

VDMPEGDecoder::~VDMPEGDecoder() {
	Shutdown();
}

IVDMPEGDecoder *CreateVDMPEGDecoder() {
	return new VDMPEGDecoder();
}

inline bool VDMPEGDecoder::mpeg_alloc() {
	unsigned long chroma_size;
	unsigned long mbw, mbh, i;

	// We need to know the following for sure:
	// pelWidth, pelHeight, and chroma_format

	if (pelWidth > 0 && pelHeight > 0) {
		mbw = (pelWidth+15)/16;
		mbh = 2*((pelHeight+31)/32);

		// size of Y buffer (3 of these)
		i = mbw*16 * mbh*16 * MPEG_BUFFERS;

		// size of U/V buffer (6 of these)
		switch (chroma_format) {
		case CHROMA420:
			chroma_size = mbw*8 * mbh*8;
			break;
		case CHROMA422:
			chroma_size = mbw*8 * mbh*16;
			break;
		case CHROMA444:
			chroma_size = mbw*16 * mbh*16;
			break;
		default:
			return false;
		}
		i += chroma_size*MPEG_BUFFERS*2;

		// Intermediate 4:2:2 buffer (1 of these)
		i += mbw*8 * mbh*16;
		// Intermediate 4:4:4 buffers (2 of these)
		i += mbw*16 * mbh*16 * 2;

		// dct coefficient block
		i += 12 * 64 * sizeof(short);

		i += 16;	// 16-byte alignment (necessary?)

// Just for laughs, here are the buffer requirements
// for SD 720 x 480 MPEG at different chroma formats:
//		CHROMA420:  1386352 bytes
//		CHROMA422:  1904752 bytes
//		CHROMA444:  2941552 bytes
// And for HDTV 1920 x 1080:
//		CHROMA420:  8362816 bytes
//		CHROMA422: 11496256 bytes
//		CHROMA444: 17763136 bytes

		memblock = VirtualAlloc(NULL, i ,MEM_COMMIT, PAGE_READWRITE);

		if (memblock != NULL) {

			// Assign memory blocks (align to 16-byte boundary)

			buffers[0].Y = (YCCSample *)
				(((unsigned long)memblock + 15) & 0xFFFFFFF0);

			for (i = 1; i < MPEG_BUFFERS; i++) {
				buffers[i].Y = buffers[i-1].Y + mbw*16 * mbh*16;
			}

			buffers[0].U = buffers[i-1].Y + mbw*16 * mbh*16;

			for (i = 1; i < MPEG_BUFFERS; i++) {
				buffers[i].U = buffers[i-1].U + chroma_size;
			}

			buffers[0].V = buffers[i-1].U + chroma_size;

			for (i = 1; i < MPEG_BUFFERS; i++) {
				buffers[i].V = buffers[i-1].V + chroma_size;
			}

			uv422 = buffers[i-1].V + chroma_size;

			u444 = uv422 + mbw*8 * mbh*16;
			v444 = u444 + mbw*16 * mbh*16;

			dct_coeff = (short (*)[64])(v444 + mbw*16 * mbh*16);

			for (i = 0; i < MPEG_BUFFERS; i++) {
				buffers[i].frame_num = -1;
			}

			return true;
		}
	}

	return false;
}

bool VDMPEGDecoder::Init(int width, int height, long seq_ext) {
	Shutdown();

// Initialize MPEG decoder fixed properties
// ----------------------------------------
// The following are considered fixed:
//		pelWidth, pelHeight
//		progressive_sequence
//		chroma_format
//		block_count
//		MPEG2_Flag
//		memblock
// These will never change during decoding.

	pelWidth = width;
	pelHeight = height;

//	current_frame = forw_frame = back_frame = (&(buffers[0]));

	if (seq_ext) {
		// Initializing MPEG-2
		// "seq_ext" is a bswapped portion of
		// the beginning of sequence_extension:
		// ........ ........ ........ 0001....	sequence_extension ID
		// ........ ........ ........ ....x...	escape bit
		// ........ ........ ........ .....xxx	profile
		// ........ ........ xxxx.... ........	level
		// ........ ........ ....x... ........	progressive_sequence
		// ........ ........ .....xx. ........	chroma_format
		// ........ x....... .......x ........	horizontal_size_ext
		// ........ .xx..... ........ ........	vertical_size_ext

		// test.mpg has 0x01005418

		//pelWidth |= ((seq_ext<<5)&0x2000)|((seq_ext>>11)&0x1000);
		//pelHeight |= (seq_ext>>9)&0x3000;
		chroma_format = (seq_ext>>9)&3;
		progressive_sequence = (seq_ext>>11)&1;

#ifdef ATIVCR_SUPPORT
		// ATI VCR file:
		if (	chroma_format != CHROMA420
			&&	chroma_format != CHROMA422
			&&	chroma_format != CHROMA444) {

			chroma_format = CHROMA420;
		}
#endif

		MPEG2_Flag = true;
	} else {
		// Initializing MPEG-1
		chroma_format = CHROMA420;
		progressive_sequence = 1;
		MPEG2_Flag = false;
	}

	switch (chroma_format) {
		case CHROMA420:	block_count = 6;	break;
		case CHROMA422:	block_count = 8;	break;
		case CHROMA444:	block_count = 12;	break;
		default: return false;		// bad chroma_format
	}

	return mpeg_alloc();
}

void VDMPEGDecoder::Shutdown() {
	if (memblock != NULL) {
		VirtualFree(memblock, 0, MEM_RELEASE);
		memblock = NULL;
	}
}

void VDMPEGDecoder::SetIntraQuantizers(const unsigned char *pMatrix) {
	if (pMatrix)
		for (int i=0; i<64; i++)
			intramatrix[zigzag[0][i]] = pMatrix[i];
	else
		memcpy(intramatrix, intramatrix_default, 64);

	memcpy(chromaintramatrix, intramatrix, 64);
}

void VDMPEGDecoder::SetNonintraQuantizers(const unsigned char *pMatrix) {
	if (pMatrix)
		for (int i=0; i<64; i++)
			nonintramatrix[zigzag[0][i]] = pMatrix[i];
	else
		memset(nonintramatrix, 16, 64);

	memcpy(chromanonintramatrix, nonintramatrix, 64);
}

void VDMPEGDecoder::SetPredictors(const VDMPEGPredictorSet *pPredictors) {
	// Won't work for MPEG-2
}
void VDMPEGDecoder::SetConverters(const VDMPEGConverterSet *pConverters) {
	// Won't work for MPEG-2
}
void VDMPEGDecoder::SetIDCTs(const VDMPEGIDCTSet *pIDCTs) {
	// Won't work with this code
}

long VDMPEGDecoder::GetErrorState() {
    return mErrorState;
}


///////////////////////////////////////////////////////////////////////////

// Critical bits-getting functions

/*
#pragma warning(push)
#pragma warning(disable:4035)

#if _MSC_VER >= 1300
	extern unsigned long _byteswap_ulong(unsigned long v);
	#pragma intrinsic(_byteswap_ulong)
	static inline unsigned long bswap(unsigned long v) {
		return _byteswap_ulong(v);
	}
#else
	static inline unsigned long bswap(unsigned long v) {
		__asm {
			mov eax,v
			bswap eax
		}
	}
#endif

#pragma warning(pop)

long VDMPEGDecoder::show_bits(int nBits) {
// Correct results are guaranteed only if nBits >= 0 and nBits <= 24
	const unsigned char *src = bitarray + (bitpos >> 3);
	unsigned long i = 0;

	// Note: ((remains + 7) >> 3) is how many BYTES remain!

	if (nBits > 0 && nBits <= 24) {
		long remains = totbits - bitpos;
		if (remains > 24) {
			// At least four bytes in buffer
			i = bswap(*((const unsigned long *)src));
		} else if (remains > 0) {
			i = src[0] << 24;
			if (remains > 8) {
				i += src[1] << 16;
				if (remains > 16) {
					i += src[2] << 8;
				}
			}
		}
		i <<= (bitpos & 7);
		i >>= (32 - nBits);
	}
	return i;
}
*/

// Well, I don't like the way VC assembles the above.
// The following hand-tweaked is about 3 times faster:

#pragma warning(push)
#pragma warning(disable:4035)	// no return value

long __declspec(naked) VDMPEGDecoder::show_bits(int nBits) {

// Correct results are guaranteed only if nBits >= 0 and nBits <= 24
// Optimize for best-case

	__asm {
		push	esi
		xor		eax, eax				; eax = return value
		push	ecx						; we'll need cl
		mov		esi, [ecx].bitpos
		mov		edx, [ecx].totbits
		cmp		dword ptr [esp+12], 0	; nBits <= 0?
		jle		done
		cmp		dword ptr [esp+12], 24	; nBits > 24?
		jg		done
		sub		edx, esi				; edx = totbits - bitpos
		jle		done					; remains <= 0 bits?
		shr		esi, 3					; byte position = bitpos / 8
		add		esi, [ecx].bitarray		; esi = bits pointer
		cmp		edx, 24
		jle		lte24					; remains <= 24 bits?
		// At least 32 bits remain in buffer (best-case)
		mov		eax, [esi]				; b3 b2 b1 b0
finale:
		mov		ecx, [ecx].bitpos
		mov		edx, 32
		bswap	eax
		sub		edx, [esp+12]			; edx = 32 - nBits
		and		ecx, 7					; ecx = bitpos & 7
		shl		eax, cl
		mov		cl, dl
		shr		eax, cl
done:
		pop		ecx
		pop		esi
		ret		4

lte24:
		cmp		edx, 16
		jle		lte16					; remains <= 16 bits?
		// Exactly 24 bits remain in buffer
		mov		al, [esi+2]
		shl		eax, 16
		mov		ax, word ptr [esi]		; 00 b2 b1 b0
		jmp		finale
lte16:
		cmp		edx, 8
		jle		lte08					; remains <= 8 bits?
		// Exactly 16 bits remain in buffer
		mov		ax, word ptr [esi]
		jmp		finale					; 00 00 b1 b0
lte08:
		test	edx, edx
		jle		finale
		// Only 8 bits remain in buffer
		mov		al, [esi]				; 00 00 00 b0
		jmp		finale
	}
}

// get_bits(x) is simply show_bits(x) + flush_buffer(x), but
// we'll avoid the overhead of a function call by copying the
// show_bits() code directly into get_bits().

long __declspec(naked) VDMPEGDecoder::get_bits(int nBits) {

// Correct results are guaranteed only if nBits >= 0 and nBits <= 24
// Optimize for best-case

	__asm {
		push	esi
		xor		eax, eax				; eax = return value
		push	ecx						; we'll need cl
		mov		esi, [ecx].bitpos
		mov		edx, [ecx].totbits
		cmp		dword ptr [esp+12], 0	; nBits <= 0?
		jle		done
		cmp		dword ptr [esp+12], 24	; nBits > 24?
		jg		done
		sub		edx, esi				; edx = totbits - bitpos
		jle		done					; remains <= 0 bits?
		shr		esi, 3					; byte position = bitpos / 8
		add		esi, [ecx].bitarray		; esi = bits pointer
		cmp		edx, 24
		jle		lte24					; remains <= 24 bits?
		// At least 32 bits remain in buffer (best-case)
		mov		eax, [esi]				; b3 b2 b1 b0
finale:
		mov		esi, [esp+12]			; esi = nBits
		mov		edx, [ecx].bitpos
		add		[ecx].bitpos, esi		; bitpos += nBits
		mov		ecx, edx
		mov		edx, 32
		bswap	eax
		sub		edx, esi				; edx = 32 - nBits
		and		ecx, 7					; ecx = bitpos & 7
		shl		eax, cl
		mov		cl, dl
		shr		eax, cl
done:
		pop		ecx
		pop		esi
		ret		4

lte24:
		cmp		edx, 16
		jle		lte16					; remains <= 16 bits?
		// Exactly 24 bits remain in buffer
		mov		al, [esi+2]
		shl		eax, 16
		mov		ax, word ptr [esi]		; 00 b2 b1 b0
		jmp		finale
lte16:
		cmp		edx, 8
		jle		lte08					; remains <= 8 bits?
		// Exactly 16 bits remain in buffer
		mov		ax, word ptr [esi]
		jmp		finale					; 00 00 b1 b0
lte08:
		test	edx, edx
		jle		finale
		// Only 8 bits remain in buffer
		mov		al, [esi]				; 00 00 00 b0
		jmp		finale
	}
}

#pragma warning(pop)		// 4035: No return value


#define flush_buffer(nBits) bitpos += nBits

/*
long VDMPEGDecoder::get_bits(int nBits) {
	long i = show_bits(nBits);
	flush_buffer(nBits);
	return i;
}
*/

int VDMPEGDecoder::next_start_code() {
	int i;

	//align to next byte
	bitpos += (-(long)bitpos) & 7;

	//set position at next start code and return it
	while (bitpos < totbits) {
		i = get_bits(8);
		if (i == 0) {
			i = show_bits(24);
			if ((i & 0xFFFFFF00) == 0x100) {
				bitpos -= 8;
				return i;
			}
		}
	}
	return 0;
}

////////////////////////////////////////////////////////////////////////////////

inline void VDMPEGDecoder::picture_coding_extension() {
	current_frame->bitflags = 0;
/*
	f_code[0][0] = get_bits(4)-1;
	f_code[0][1] = get_bits(4)-1;
	f_code[1][0] = get_bits(4)-1;
	f_code[1][1] = get_bits(4)-1;

	// zero is forbidden
	if (f_code[0][0] < 0) f_code[0][0] = 0;
	if (f_code[0][1] < 0) f_code[0][1] = 0;
	if (f_code[1][0] < 0) f_code[1][0] = 0;
	if (f_code[1][1] < 0) f_code[1][1] = 0;
*/
	if (f_code[0][0] = get_bits(4)) f_code[0][0]--;
	if (f_code[0][1] = get_bits(4)) f_code[0][1]--;
	if (f_code[1][0] = get_bits(4)) f_code[1][0]--;
	if (f_code[1][1] = get_bits(4)) f_code[1][1]--;

	intra_dc_precision = get_bits(2);
	picture_structure = get_bits(2);
	if (get_bits(1)) current_frame->bitflags |= PICTUREFLAG_TFF;
	frame_pred_frame_dct = get_bits(1);
	concealment_vectors = get_bits(1);
	q_scale_type = get_bits(1);
	intra_vlc_format = get_bits(1);
	alternate_scan = get_bits(1);
	if (get_bits(1)) current_frame->bitflags |= PICTUREFLAG_RFF;
	flush_buffer(1);	// chroma_420_type (obsolete)
	if (get_bits(1)) current_frame->bitflags |= PICTUREFLAG_PF;
}

inline void VDMPEGDecoder::quant_matrix_extension() {
	if (get_bits(1)) {
		for (int i=0; i<64; ++i)
			intramatrix[zigzag[0][i]] = (unsigned char)get_bits(8);
	}

	if (get_bits(1)) {
		for (int i=0; i<64; ++i)
			nonintramatrix[zigzag[0][i]] = (unsigned char)get_bits(8);
	}

	if (get_bits(1)) {
		for (int i=0; i<64; ++i)
			chromaintramatrix[zigzag[0][i]] = (unsigned char)get_bits(8);
	} else
		memcpy(chromaintramatrix, intramatrix, 64);

	if (get_bits(1)) {
		for (int i=0; i<64; ++i)
			chromanonintramatrix[zigzag[0][i]] = (unsigned char)get_bits(8);
	} else
		memcpy(chromanonintramatrix, nonintramatrix, 64);
}

void VDMPEGDecoder::extension_and_user_data() {
	int code;

	while (code = next_start_code()) {
		if (code == EXTENSION_START_CODE) {
			flush_buffer(32);
			switch (get_bits(4)) {

			case SEQUENCE_EXTENSION_ID:
				flush_buffer(8+7+12+1+8+1+2+5);
				if (fAllowMatrixCoefficients) {
					matrix_coefficients = 1;
				}
				break;

			case SEQUENCE_DISPLAY_EXTENSION_ID:
				if (get_bits(4) & 1) {
					flush_buffer(16);	// color_primaries, etc.
					if (fAllowMatrixCoefficients) {
						matrix_coefficients = get_bits(8) & 7;
					} else {
						flush_buffer(8);
					}
				}
				break;

			case PICTURE_CODING_EXTENSION_ID:
				picture_coding_extension();
				break;

			case QUANT_MATRIX_EXTENSION_ID:
				quant_matrix_extension();
				break;
			}

		} else {
			if (code != USER_DATA_START_CODE) break;
			flush_buffer(32);
		}
	}
}

inline bool VDMPEGDecoder::sequence_header() {

	if (get_bits(12) != (pelWidth & 0xFFF)) return false;
	if (get_bits(12) != (pelHeight & 0xFFF)) return false;
	flush_buffer(4);	// aspect_ratio_code
	flush_buffer(4);	// frame_rate_code
	flush_buffer(18);	// bit_rate
	flush_buffer(1);	// marker_bit
	flush_buffer(10);	// vbv_buffer_size
	flush_buffer(1);	// constrained_parameters_flag

	// load new matrices if applicable
	if (get_bits(1)) {	// load_intra_quantizer_matrix
		for (int i = 0; i < 64; ++i) {
			intramatrix[zigzag[0][i]] = (unsigned char)get_bits(8);
		}
	} else {
		memcpy(intramatrix, intramatrix_default, 64);
	}

	if (get_bits(1)) {	// load_non_intra_quantizer_matrix
		for (int i = 0; i < 64; ++i) {
			nonintramatrix[zigzag[0][i]] = (unsigned char)get_bits(8);
		}
	} else {
		memset(nonintramatrix, 16, 64);
	}

	memcpy(chromaintramatrix, intramatrix, 64);
	memcpy(chromanonintramatrix, nonintramatrix, 64);

	matrix_coefficients = 5;
	extension_and_user_data();

	return true;
}

inline void VDMPEGDecoder::picture_header() {
	flush_buffer(10);	// temporal_reference
	frame_type = get_bits(3);

	flush_buffer(16);	// VBV_delay

	if (frame_type == MPEG_FRAME_TYPE_P || frame_type == MPEG_FRAME_TYPE_B) {
		forw_vector_full_pel = get_bits(1);
		f_code[0][0] = f_code[0][1] = get_bits(3)-1;

		if (frame_type == MPEG_FRAME_TYPE_B) {
			back_vector_full_pel = get_bits(1);
			f_code[1][0] = f_code[1][1] = get_bits(3)-1;
		}
	}

	// The following remain at MPEG-1 defaults, unless
	// overwritten by MPEG-2 picture_coding_extension()...
	picture_structure = FRAME_PICTURE;
	frame_pred_frame_dct = 1;
	concealment_vectors = 0;
	current_frame->bitflags = PICTUREFLAG_PF;

	extension_and_user_data();
}

// frame is the current frame number
// dst is destination MPEGBuffer
// fwd is forward reference MPEGBuffer
// rev is backward reference MPEGBuffer
int VDMPEGDecoder::DecodeFrame(const void *src, long len, long frame, int dst, int fwd, int rev) {
	int code;

// sanity checks
	if (memblock == NULL) return -1;
	if ((src == NULL) || (len < 4)) return -1;
	if ((unsigned int)dst >= MPEG_BUFFERS) return -1;
	if ((unsigned int)fwd >= MPEG_BUFFERS) return -1;
	if ((unsigned int)rev >= MPEG_BUFFERS) return -1;

// Bind bits-getting functions to "input_data"
	bitarray = (unsigned char *)src;
	totbits = len << 3;
	bitpos = 0;
	second_field = 0;
	frame_type = 0;

	// Note:  The confusion here stems from the fact that the most
	// recently decoded I/P frame is the _forward_ prediction for
	// a P frame, but it's the _backward_ prediction for a B frame!

	// Rule:  The most recently decoded I/P frame is
	// always found in the _backward_ prediction buffer

	forw_frame = &(buffers[fwd]);
	back_frame = &(buffers[rev]);
	current_frame = &(buffers[dst]);

Again:

// Decode headers up to the first slice
	while (code = next_start_code()) {
		if (code >= SLICE_START_CODE_MIN
			&& code <= SLICE_START_CODE_MAX) break;

		flush_buffer(32);

		if (code == SEQUENCE_START_CODE) {
			if (!sequence_header()) {
				code = 0;
				break;
			}
		} else if (code == PICTURE_START_CODE) {
			picture_header();
		}
	}

	if (!code) return -1;	// error; no slices

	mbWidth = (pelWidth+15)>>4;
	if (progressive_sequence) {
		mbHeight = (pelHeight+15)>>4;
	} else {
		mbHeight = (pelHeight+31)>>5;
		if (picture_structure == FRAME_PICTURE) mbHeight<<=1;
	}
	y_pitch = mbWidth << 4;

// Decode all slices
	do {
		flush_buffer(32);
		slice(code);
		code = next_start_code();
	} while (code >= SLICE_START_CODE_MIN
		&& code <= SLICE_START_CODE_MAX);

	if (!second_field) {
		current_frame->frame_num = frame;

		if (picture_structure != FRAME_PICTURE) {
			second_field = !second_field;
			goto Again;
		}
	}

	if (cpuflags & CPUF_SUPPORTS_MMX) {
		if (cpuflags & CPUF_SUPPORTS_INTEGER_SSE) __asm sfence
		__asm emms
	}

	return dst;
}

int VDMPEGDecoder::GetFrameBuffer(long frame) {
	int i;
	for (i = 0; i < MPEG_BUFFERS; ++i) {
		if (buffers[i].frame_num == frame) return i;
	}
	return -1;
}

long VDMPEGDecoder::GetFrameNumber(int buffer) {
	if ((unsigned int)buffer < MPEG_BUFFERS) {
		return buffers[buffer].frame_num;
	}
	return -1;
}

void VDMPEGDecoder::CopyFrameBuffer(int dst, int src, long frameno) {
	if (memblock != NULL && (unsigned int)dst < MPEG_BUFFERS && (unsigned int)src < MPEG_BUFFERS) {
		VDMemcpyRect(buffers[dst].Y, y_pitch, buffers[src].Y, y_pitch, mbWidth * 16, mbHeight * 16);

		switch (chroma_format) {
		case CHROMA420:
			VDMemcpyRect(buffers[dst].U, y_pitch >> 1, buffers[src].U, y_pitch >> 1, mbWidth * 8, mbHeight * 8);
			VDMemcpyRect(buffers[dst].V, y_pitch >> 1, buffers[src].V, y_pitch >> 1, mbWidth * 8, mbHeight * 8);
		case CHROMA422:
			VDMemcpyRect(buffers[dst].U, y_pitch, buffers[src].U, y_pitch, mbWidth * 16, mbHeight * 8);
			VDMemcpyRect(buffers[dst].V, y_pitch, buffers[src].V, y_pitch, mbWidth * 16, mbHeight * 8);
		case CHROMA444:
			VDMemcpyRect(buffers[dst].U, y_pitch, buffers[src].U, y_pitch, mbWidth * 16, mbHeight * 16);
			VDMemcpyRect(buffers[dst].V, y_pitch, buffers[src].V, y_pitch, mbWidth * 16, mbHeight * 16);
		}

		buffers[dst].frame_num = frameno;
		buffers[dst].bitflags = buffers[src].bitflags;
	}
}

void VDMPEGDecoder::SwapFrameBuffers(int dst, int src) {
	if (	(unsigned int)dst < MPEG_BUFFERS
		&&	(unsigned int)src < MPEG_BUFFERS
		&&	dst != src) {

		MPEGBuffer b = buffers[dst];
		buffers[dst] = buffers[src];
		buffers[src] = b;
	}
}

void VDMPEGDecoder::ClearFrameBuffers() {
	int i;
	for (i = 0; i < MPEG_BUFFERS; ++i) {
		buffers[i].frame_num = -1;
	}
}

void VDMPEGDecoder::CopyField(int dst, int dstfield, int src, int srcfield) {
	if (memblock != NULL
		&&	(unsigned int)dst < MPEG_BUFFERS
		&&	(unsigned int)src < MPEG_BUFFERS) {

	// Copy a single field from src to dst

		YCCSample *srcf;
		YCCSample *dstf;
		long cpw, cph, h;

		cpw = (pelWidth + 15) & -16;
		cph = (pelHeight + 15) & -16;

		cph >>= 1;

		if (srcfield) srcfield = cpw;
		if (dstfield) dstfield = cpw;

	// First, the Y
		srcf = buffers[src].Y + srcfield;	// source field
		dstf = buffers[dst].Y + dstfield;	// destination field

	// No need to copy anything if they are the same
		if (dstf == srcf) return;

		for (h = 0; h < cph; ++h) {
			memcpy(dstf, srcf, cpw);
			dstf += cpw * 2;
			srcf += cpw * 2;
		}

	// Next, U and V
		if (chroma_format != CHROMA444) {
			srcfield >>= 1;
			dstfield >>= 1;
			cpw >>= 1;
			if (chroma_format == CHROMA420) {
				cph >>= 1;
			}
		}

		srcf = buffers[src].U + srcfield;
		dstf = buffers[dst].U + dstfield;

		for (h = 0; h < cph; ++h) {
			memcpy(dstf, srcf, cpw);
			dstf += cpw * 2;
			srcf += cpw * 2;
		}

		srcf = buffers[src].V + srcfield;
		dstf = buffers[dst].V + dstfield;

		for (h = 0; h < cph; ++h) {
			memcpy(dstf, srcf, cpw);
			dstf += cpw * 2;
			srcf += cpw * 2;
		}

		// Invalidate this frame number,
		// and flag the picture as interlaced
		buffers[dst].frame_num = -1;
		buffers[dst].bitflags &= (~PICTUREFLAG_PF);
	}
}

/*********************************************************
	Integer IDCT, based on Chen-Wang algorithm
**********************************************************/

#define CW_W1	2841	// 2048*sqrt(2)*cos(1*pi/16)
#define CW_W2	2676	// 2048*sqrt(2)*cos(2*pi/16)
#define CW_W3	2408	// 2048*sqrt(2)*cos(3*pi/16)
#define CW_W5	1609	// 2048*sqrt(2)*cos(5*pi/16)
#define CW_W6	1108	// 2048*sqrt(2)*cos(6*pi/16)
#define CW_W7	565		// 2048*sqrt(2)*cos(7*pi/16)

inline void VDMPEGDecoder::idctrow(short *blk) {
	long x0,x1,x2,x3,x4,x5,x6,x7,x8;

	// pos is (col*8+#)
	x1 = blk[4]<<11;
	x2 = blk[6];
	x3 = blk[2];
	x4 = blk[1];
	x5 = blk[7];
	x6 = blk[5];
	x7 = blk[3];

	if (!(x1|x2|x3|x4|x5|x6|x7)) {
		// Shortcut...
		x0 = blk[0]<<3;
		blk[0]=blk[1]=blk[2]=blk[3]
			=blk[4]=blk[5]=blk[6]=blk[7]=(short)x0;
		return;
	}

	x0=(blk[0]<<11)+128;    // for proper rounding

	// first stage
	x8 = CW_W7 * (x4 + x5);
	x4 = x8 + (CW_W1 - CW_W7) * x4;
	x5 = x8 - (CW_W1 + CW_W7) * x5;
	x8 = CW_W3 * (x6 + x7);
	x6 = x8 - (CW_W3 - CW_W5) * x6;
	x7 = x8 - (CW_W3 + CW_W5) * x7;
	x8 = CW_W6 * (x3 + x2);
	x2 = x8 - (CW_W2 + CW_W6) * x2;
	x3 = x8 + (CW_W2 - CW_W6) * x3;

	// second stage
	x8 = x0 + x1; x0 -= x1;
	x1 = x4 + x6; x4 -= x6;
	x6 = x5 + x7; x5 -= x7;
	x7 = x8 + x3; x8 -= x3;
	x3 = x0 + x2; x0 -= x2;
	x2 = (181 * (x4 + x5) + 128)>>8;
	x4 = (181 * (x4 - x5) + 128)>>8;

	// final stage
	blk[0] = (short)((x7 + x1)>>8);
	blk[1] = (short)((x3 + x2)>>8);
	blk[2] = (short)((x0 + x4)>>8);
	blk[3] = (short)((x8 + x6)>>8);
	blk[4] = (short)((x8 - x6)>>8);
	blk[5] = (short)((x0 - x4)>>8);
	blk[6] = (short)((x3 - x2)>>8);
	blk[7] = (short)((x7 - x1)>>8);
}

inline void VDMPEGDecoder::idctcol(short *blk) {
	long x0,x1,x2,x3,x4,x5,x6,x7,x8;

	x4 = blk[8];
	x3 = blk[16];
	x7 = blk[24];
	x1 = blk[32]<<8;
	x6 = blk[40];
	x2 = blk[48];
	x5 = blk[56];

	if (!(x1|x2|x3|x4|x5|x6|x7)) {
		// Shortcut...
		x0 = (blk[0]+32) >> 6;
		if (x0 < -256) x0 = -256; else if (x0 > 255) x0 = 255;
		blk[0]=blk[8]=blk[16]=blk[24]=blk[32]
			=blk[40]=blk[48]=blk[56] = (short)x0;
		return;
	}

	x0 = (blk[0] << 8) + 8192;

	// first stage
	x8 = (CW_W7 * (x4 + x5)) + 4;
	x4 = (x8 + (CW_W1 - CW_W7) * x4) >> 3;
	x5 = (x8 - (CW_W1 + CW_W7) * x5) >> 3;
	x8 = (CW_W3 * (x6 + x7)) + 4;
	x6 = (x8 - (CW_W3 - CW_W5) * x6) >> 3;
	x7 = (x8 - (CW_W3 + CW_W5) * x7) >> 3;
	x8 = (CW_W6 * (x3 + x2)) + 4;
	x2 = (x8 - (CW_W2 + CW_W6) * x2) >> 3;
	x3 = (x8 + (CW_W2 - CW_W6) * x3) >> 3;

	// second stage
	x8 = x0 + x1; x0 -= x1;
	x1 = x4 + x6; x4 -= x6;
	x6 = x5 + x7; x5 -= x7;
	x7 = x8 + x3; x8 -= x3;
	x3 = x0 + x2; x0 -= x2;
	x2 = (181 * (x4 + x5) + 128) >> 8;
	x4 = (181 * (x4 - x5) + 128) >> 8;

	// final stage
	x5 = (x7 + x1) >> 14;
	if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
	blk[0] = (short)x5;
	x5 = (x3 + x2) >> 14;
	if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
	blk[8] = (short)x5;
	x5 = (x0 + x4) >> 14;
	if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
	blk[16] = (short)x5;
	x5 = (x8 + x6) >> 14;
	if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
	blk[24] = (short)x5;
	x5 = (x8 - x6) >> 14;
	if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
	blk[32] = (short)x5;
	x5 = (x0 - x4) >> 14;
	if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
	blk[40] = (short)x5;
	x5 = (x3 - x2) >> 14;
	if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
	blk[48] = (short)x5;
	x5 = (x7 - x1) >> 14;
	if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
	blk[56] = (short)x5;
}

void VDMPEGDecoder::Fast_IDCT(void *src) {
	int i;
	short *blk;
// two dimensional inverse discrete cosine transform
	for (i=0,blk=(short *)src; i<8; ++i,blk+=8) idctrow(blk);
	for (i=0,blk=(short *)src; i<8; ++i,++blk) idctcol(blk);
}

void VDMPEGDecoder::Add_Block(int comp, int bx, int by, int addflag) {
	int i,j,v,iincr,uv_pitch;
	YCCSample *rfp;

	if (comp < 4) {     // luminance
		rfp = current_frame->Y;
		if (picture_structure==FRAME_PICTURE) {
			if (dct_type) {
				// field DCT coding
				rfp += y_pitch*(by+((comp&2)>>1)) + bx + ((comp&1)<<3);
				iincr = y_pitch<<1;
			} else {
				// frame DCT coding
				rfp += y_pitch*(by+((comp&2)<<2)) + bx + ((comp&1)<<3);
				iincr = y_pitch;
			}
		} else {
			// field picture
			if (picture_structure == BOTTOM_FIELD) rfp += y_pitch;
			rfp += (y_pitch<<1)*(by+((comp&2)<<2)) + bx + ((comp&1)<<3);
			iincr = y_pitch<<1;
		}
	} else {        // chrominance
		if (comp & 1)	rfp = current_frame->V;
		else			rfp = current_frame->U;

		// scale coordinates
		if (chroma_format != CHROMA444) {
			bx >>= 1;
			if (chroma_format == CHROMA420) by >>= 1;
			uv_pitch = y_pitch >> 1;
		} else {
			uv_pitch = y_pitch;
		}

		if (picture_structure==FRAME_PICTURE) {
			if (dct_type && (chroma_format != CHROMA420)) {
				// field DCT coding
				rfp += uv_pitch*(by+((comp&2)>>1)) + bx + (comp&8);
				iincr = uv_pitch<<1;
			} else {
				// frame DCT coding
				rfp += uv_pitch*(by+((comp&2)<<2)) + bx + (comp&8);
				iincr = uv_pitch;
			}
		} else {
			// field picture
			if (picture_structure == BOTTOM_FIELD) rfp += uv_pitch;
			rfp += (uv_pitch<<1)*(by+((comp&2)<<2)) + bx + (comp&8);
			iincr = uv_pitch<<1;
		}
	}

	if (cpuflags & CPUF_SUPPORTS_MMX) {
		mmx_Add_Block(dct_coeff[comp], rfp, iincr, addflag);

	} else {
		short *blk = dct_coeff[comp];

		if (addflag) {
			for (j=0; j<8; ++j) {
				for (i=0; i<8; ++i) {
					v = *blk++ + rfp[i];
					if ((unsigned int)v > 255) v = (~v >> 31) & 255;
					rfp[i] = (YCCSample)v;
				}
				rfp+= iincr;
			}
		} else {
			for (j=0; j<8; ++j) {
				for (i=0; i<8; ++i) {
					v = *blk++ + 128;
					if ((unsigned int)v > 255) v = (~v >> 31) & 255;
					rfp[i] = (YCCSample)v;
				}
				rfp+= iincr;
			}
		}
	}
}

long VDMPEGDecoder::get_luma_diff() {
	int size;
	// decode length
	int code = show_bits(9);	// 1 1111 ....

	if (code < 0x1F0) {
		code >>= 4;
		size = DClumtab0[code].val;
		flush_buffer(DClumtab0[code].len);
	} else {
		code &= 0xF;
		size = DClumtab1[code].val;
		flush_buffer(DClumtab1[code].len);
	}

	if (size > 0) {
		code = get_bits(size);	// maximum is 11 bits
		size = 1 << (size-1);
		if (code < size) code = (code+1) - (size+size);
	} else code = 0;

	return code;
}

long VDMPEGDecoder::get_chroma_diff() {
	int size;
	// decode length
	int code = show_bits(10);

	if (code < 0x3E0) {
		code >>= 5;
		size = DCchromtab0[code].val;
		flush_buffer(DCchromtab0[code].len);
	} else {
		code &= 0x1F;
		size = DCchromtab1[code].val;
		flush_buffer(DCchromtab1[code].len);
	}

	if (size > 0) {
		code = get_bits(size);	// maximum is 11 bits
		size = 1 << (size-1);
		if (code < size) code = (code+1) - (size+size);
	} else code = 0;

	return code;
}

// decode one intra coded MPEG-1 block
void VDMPEGDecoder::Decode_MPEG1_Intra_Block(int comp) {
	long code,val,i,j,sign;
	const DCTtab *tab;

	short *blk = dct_coeff[comp];

	// decode DC coefficients
	if (comp < 4)		// Y
		val = (dct_dc_y_past += get_luma_diff());
	else if (comp & 1)	// V
		val = (dct_dc_v_past += get_chroma_diff());
	else				// U
		val = (dct_dc_u_past += get_chroma_diff());

	blk[0] = (short)(val << 3);

	if (frame_type == MPEG_FRAME_TYPE_D) return;

	// decode AC coefficients
	for (i=1; ; ++i) {
		code = show_bits(16);
		if (code>=16384)
			tab = &DCTtabnext[(code>>12)-4];
		else if (code>=1024)
			tab = &DCTtab0[(code>>8)-4];
		else if (code>=512)
			tab = &DCTtab1[(code>>6)-8];
		else if (code>=256)
			tab = &DCTtab2[(code>>4)-16];
		else if (code>=128)
			tab = &DCTtab3[(code>>3)-16];
		else if (code>=64)
			tab = &DCTtab4[(code>>2)-16];
		else if (code>=32)
			tab = &DCTtab5[(code>>1)-16];
		else if (code>=16)
			tab = &DCTtab6[code-16];
		else {
			SetError(kErrorBadValue);
			break;
		}

		flush_buffer(tab->len);
		val = tab->run;

		if (val == 65) {	// escape
			i += get_bits(6);

			val = get_bits(8);
			if (val==0)
				val = get_bits(8);
			else if (val==128)
				val = get_bits(8) - 256;
			else if (val>128)
				val -= 256;

			if ((sign = (val<0))) val = -val;
		} else {
			if (val == 64) break;
			i+= val;
			val = tab->level;
			sign = get_bits(1);
		}
		if (i>=64) {
			SetError(kErrorTooManyCoefficients);
			break;
		}

		j = zigzag[0][i];
		val = (val * quantizer_scale * intramatrix[j]) >> 3;
		if (val) val = (val - 1) | 1;		// mismatch
		if (val >= 2048) val = 2047 + sign;	// saturation
		if (sign) val = -val;
		blk[j] = (short)val;
    }
}

// decode one non-intra coded MPEG-1 block
inline void VDMPEGDecoder::Decode_MPEG1_Non_Intra_Block(int comp) {
	long code,val,i,j,sign;
	const DCTtab *tab;

	short *blk = dct_coeff[comp];

	// decode AC coefficients
	for (i=0; ; ++i) {
		code = show_bits(16);
		if (code>=16384)
			if (i)	tab = &DCTtabnext[(code>>12)-4];
			else	tab = &DCTtabfirst[(code>>12)-4];
		else if (code>=1024)
			tab = &DCTtab0[(code>>8)-4];
		else if (code>=512)
			tab = &DCTtab1[(code>>6)-8];
		else if (code>=256)
			tab = &DCTtab2[(code>>4)-16];
		else if (code>=128)
			tab = &DCTtab3[(code>>3)-16];
		else if (code>=64)
			tab = &DCTtab4[(code>>2)-16];
		else if (code>=32)
			tab = &DCTtab5[(code>>1)-16];
		else if (code>=16)
			tab = &DCTtab6[code-16];
		else {
			SetError(kErrorBadValue);
			break;
		}

		flush_buffer(tab->len);
		val = tab->run;

		if (val == 65) {	// escape
			i+= get_bits(6);

			val = get_bits(8);
			if (val==0)
				val = get_bits(8);
			else if (val==128)
				val = get_bits(8) - 256;
			else if (val>128)
				val -= 256;

			if ((sign = (val<0))) val = -val;
		} else {
			if (val == 64) break;
			i+= val;
			val = tab->level;
			sign = get_bits(1);
		}
		if (i >= 64) {
			SetError(kErrorTooManyCoefficients);
			break;
		}

		j = zigzag[0][i];
		val = (((val<<1)+1) * quantizer_scale * nonintramatrix[j]) >> 4;
		if (val) val = (val - 1) | 1;		// mismatch
		if (val >= 2048) val = 2047 + sign;	// saturation
		if (sign) val = -val;
		blk[j] = (short)val;
	}
}

// decode one intra coded MPEG-2 block
inline void VDMPEGDecoder::Decode_MPEG2_Intra_Block(int comp) {
	long code,val,i,j,sign,sum;
	const DCTtab *tab;

	short *blk = dct_coeff[comp];

	unsigned char *qmat = (comp<4 || chroma_format==CHROMA420)?
		intramatrix: chromaintramatrix;

	// DC coefficient
	if (comp < 4)		// Y
		val = (dct_dc_y_past += get_luma_diff());
	else if (comp & 1)	// V
		val = (dct_dc_v_past += get_chroma_diff());
	else				// U
		val = (dct_dc_u_past += get_chroma_diff());

	sum = val << (3 - intra_dc_precision);
	blk[0] = (short)sum;

	// AC coefficients
	for (i=1; ; ++i) {
		code = show_bits(16);
		if (code>=16384)
			if (intra_vlc_format)	tab = &DCTtab0a[(code>>8)-4];
			else					tab = &DCTtabnext[(code>>12)-4];
		else if (code>=1024)
			if (intra_vlc_format)	tab = &DCTtab0a[(code>>8)-4];
			else					tab = &DCTtab0[(code>>8)-4];
		else if (code>=512)
			if (intra_vlc_format)	tab = &DCTtab1a[(code>>6)-8];
			else					tab = &DCTtab1[(code>>6)-8];
		else if (code>=256)
			tab = &DCTtab2[(code>>4)-16];
		else if (code>=128)
			tab = &DCTtab3[(code>>3)-16];
		else if (code>=64)
			tab = &DCTtab4[(code>>2)-16];
		else if (code>=32)
			tab = &DCTtab5[(code>>1)-16];
		else if (code>=16)
			tab = &DCTtab6[code-16];
		else {
			SetError(kErrorBadValue);
			break;
		}

		flush_buffer(tab->len);
		val = tab->run;

		if (val == 65) {	// escape
			i += get_bits(6);
			val = get_bits(12);
			if (!(val & 2047)) {
				SetError(kErrorBadValue);
				break;
			}
			if ((sign = (val >= 2048))) val = 4096 - val;
		} else {
			if (val == 64) break;
			i += val;
			val = tab->level;
			sign = get_bits(1);
		}
		if (i >= 64) {
			SetError(kErrorTooManyCoefficients);
			break;
		}

		j = zigzag[alternate_scan][i];
		val = (val * quantizer_scale * qmat[j]) >> 4;
		if (val >= 2048) val = 2047 + sign;	// saturation
		if (sign) val = -val;
		blk[j] = (short)val;
		sum ^= val;		// mismatch
	}

	if (!mErrorState)
		if (!(sum & 1)) blk[63] ^= 1;	// mismatch control
}

// decode one non-intra coded MPEG-2 block
inline void VDMPEGDecoder::Decode_MPEG2_Non_Intra_Block(int comp) {
	long code,val,i,j,sign,sum;
	const DCTtab *tab;

	short *blk = dct_coeff[comp];

	unsigned char *qmat = (comp<4 || chroma_format==CHROMA420)?
		nonintramatrix: chromanonintramatrix;

	// AC coefficients
	sum = 0;
	for (i=0; ; ++i) {
		code = show_bits(16);
		if (code>=16384)
			if (i)	tab = &DCTtabnext[(code>>12)-4];
			else	tab = &DCTtabfirst[(code>>12)-4];
		else if (code>=1024)
			tab = &DCTtab0[(code>>8)-4];
		else if (code>=512)
			tab = &DCTtab1[(code>>6)-8];
		else if (code>=256)
			tab = &DCTtab2[(code>>4)-16];
		else if (code>=128)
			tab = &DCTtab3[(code>>3)-16];
		else if (code>=64)
			tab = &DCTtab4[(code>>2)-16];
		else if (code>=32)
			tab = &DCTtab5[(code>>1)-16];
		else if (code>=16)
			tab = &DCTtab6[code-16];
		else {
			SetError(kErrorBadValue);
			break;
		}

		flush_buffer(tab->len);
		val = tab->run;

		if (val == 65) {	// escape
			i += get_bits(6);
			val = get_bits(12);
			if (!(val & 2047)) {
				SetError(kErrorBadValue);
				break;
			}
			if ((sign = (val >= 2048))) val = 4096 - val;
		} else {
			if (val == 64) break;
			i += val;
			val = tab->level;
			sign = get_bits(1);
		}
		if (i >= 64) {
			SetError(kErrorTooManyCoefficients);
			break;
		}

		j = zigzag[alternate_scan][i];
		val = (((val<<1)+1) * quantizer_scale * qmat[j]) >> 5;
		if (val >= 2048) val = 2047 + sign;	// saturation
		if (sign) val = -val;
		blk[j] = (short)val;
		sum ^= val;		// mismatch
	}

	if (!mErrorState)
		if (!(sum & 1)) blk[63] ^= 1;	// mismatch control
}


///////////////////////////
// Prediction (replaces a_predict* code)

void VDMPEGDecoder::form_component_prediction(
	YCCSample* src, YCCSample* dst, int lx, int w, int h,
	int x, int y, int dx, int dy, int average_flag) {

// src is a YCbCr field from where the prediction will be read.
// dst is a YCbCr field where the prediction will be written.

// average_flag is used for bidirectional prediction.  If average_flag
// is nonzero, then we blend (50/50) the new prediction with a previous
// prediction which is already in the destination macroblock.

// lx is the horizontal pitch (in pels) of both src and dst.
// (x, y) is the position of the current macroblock (in pels).
// (dx, dy) is a vector to the src prediction (in half pels).

// Calculate the macroblock positions in src and dst:

	src += lx * (y + (dy >> 1)) + x + (dx >> 1);
	dst += lx * y + x;

// Now bounds check the prediction.  Only the source macroblock
// needs to be checked; the destination is always in range.

// Our 3 YCbCr buffers span the range from "memblock" to "uv422",
// so anything outside that range is an out of bounds prediction.

// "src" always increments, so as long as it starts out greater
// than or equal to "memblock" we're safe.  It may increment beyond
// "uv422", but we're still safe because there is PLENTY of allocated
// memory up there, and it will never creep farther than 32 * Y_pitch
// into that area.  Note: data at "src" is only read, never written.

// (The only way the above can happen is if we're passed an out of
// bounds motion vector.  So if that's the case, we don't care about
// "perfect" decoding anymore, we just care about NOT crashing!)

	if (((DWORD)src >= (DWORD)memblock) && ((DWORD)src < (DWORD)uv422)) {

		if (cpuflags & CPUF_SUPPORTS_MMX) {

			if (cpuflags & CPUF_SUPPORTS_INTEGER_SSE) {
				isse_form_component_prediction(src, dst,
					lx, w, h, dx, dy, average_flag);

			} else {
				mmx_form_component_prediction(src, dst,
					lx, w, h, dx, dy, average_flag);
			}

		} else {

			int i, j;

/*	The following are equivalent:

a)	result = (s1 + s2 + 1) >> 1;
b)	result = (s1 >> 1) + (s2 >> 1) + ((s1 | s2) & 1);

a)	result = (d + ((s1 + s2 + 1) >> 1) + 1) >> 1;
b)	result = (s1 + s2 + d + d + 3) >> 2;

a)	result = (d + ((s1 + s2 + s3 + s4 + 2) >> 2) + 1) >> 1;
b)	result = (s1 + s2 + s3 + s4 + d + d + d + d + 6) >> 3;
*/

			if (dx & 1) {
				if (dy & 1) {
					// horizontal and vertical half-pel
					if (average_flag) {
						for (j = 0; j < h; ++j) {
							for (i = 0; i < w; ++i) {
								int v = src[i] + src[i + 1]
									+ src[i + lx] + src[i + lx + 1];

								v = dst[i] + ((v + 2) >> 2);
								dst[i] = (YCCSample)((v + 1) >> 1);
							}
							src += lx;
							dst += lx;
						}
					} else {
						for (j = 0; j < h; ++j) {
							for (i = 0; i < w; ++i) {
								int v = src[i] + src[i + 1]
									+ src[i + lx] + src[i + lx + 1];

								dst[i] = (YCCSample)((v + 2) >> 2);
							}
							src += lx;
							dst += lx;
						}
					}
				} else {
					// horizontal but no vertical half-pel
					if (average_flag) {
						for (j = 0; j < h; ++j) {
							for (i = 0; i < w; ++i) {
								int v = src[i] + src[i + 1];
								v = dst[i] + ((v + 1) >> 1);
								dst[i] = (YCCSample)((v + 1) >> 1);
							}
							src += lx;
							dst += lx;
						}
					} else {
						for (j = 0; j < h; ++j) {
							for (i = 0; i < w; ++i) {
								int v = src[i] + src[i + 1];
								dst[i] = (YCCSample)((v + 1) >> 1);
							}
							src += lx;
							dst += lx;
						}
					}
				}
			} else if (dy & 1) {
				// no horizontal but vertical half-pel
				if (average_flag) {
					for (j = 0; j < h; ++j) {
						for (i = 0; i < w; ++i) {
							int v = src[i] + src[i + lx];
							v = dst[i] + ((v + 1) >> 1);
							dst[i] = (YCCSample)((v + 1) >> 1);
						}
						src += lx;
						dst += lx;
					}
				} else {
					for (j = 0; j < h; ++j) {
						for (i = 0; i < w; ++i) {
							int v = src[i] + src[i + lx];
							dst[i] = (YCCSample)((v + 1) >> 1);
						}
						src += lx;
						dst += lx;
					}
				}
			} else {
				// no horizontal nor vertical half-pel
				if (average_flag) {
					for (j = 0; j < h; ++j) {
						for (i = 0; i < w; ++i) {
							int v = dst[i] + src[i];
							dst[i] = (YCCSample)((v + 1) >> 1);
						}
						src += lx;
						dst += lx;
					}
				} else {
					for (j = 0; j < h; ++j) {
						for (i = 0; i < w; ++i) {
							dst[i] = src[i];
						}
						src += lx;
						dst += lx;
					}
				}
			}
		}
	} else {
		SetError(kErrorBadMotionVector);
	}
}

void VDMPEGDecoder::form_prediction(
	MPEGBuffer *srcBuf, int sfield, MPEGBuffer *dstBuf, int dfield,
	int lx, int w, int h, int x, int y, int dx, int dy, int average_flag)
{

// Note: lx is always y_pitch or y_pitch*2
// Note: h is always 8 or 16
// Note: w is always 16


// Distance to the next field (if applicable):
	if (sfield) sfield = lx >> 1;
	if (dfield) dfield = lx >> 1;

// Get the Y component prediction
	form_component_prediction(
		srcBuf->Y+sfield,dstBuf->Y+dfield,
		lx,w,h,x,y,dx,dy,average_flag);

// Adjust the parameters for lower resolution chroma
	if (chroma_format != CHROMA444) {
		// 4:2:2 or 4:2:0
		sfield >>= 1;
		dfield >>= 1;	// distance to next field is halved
		lx >>= 1;		// horizontal pitch is halved
		w >>= 1;		// width of prediction is halved
		x >>= 1;		// horizontal macroblock position is halved
		dx /= 2;		// horizontal motion vector is halved

		if (chroma_format == CHROMA420) {
			// 4:2:0
			h >>= 1;	// height of prediction is halved
			y >>= 1;	// vertical macroblock position is halved
			dy /= 2;	// vertical motion vector is halved
		}
	}

// Get the Cb component prediction
	form_component_prediction(
		srcBuf->U+sfield,dstBuf->U+dfield,
		lx,w,h,x,y,dx,dy,average_flag);

// Get the Cr component prediction
	form_component_prediction(
		srcBuf->V+sfield,dstBuf->V+dfield,
		lx,w,h,x,y,dx,dy,average_flag);
}

void VDMPEGDecoder::dual_prime(int DMV[][2], int mvx, int mvy) {
	if (picture_structure == FRAME_PICTURE) {
		if (current_frame->bitflags & PICTUREFLAG_TFF) {
			DMV[0][0] = ((mvx  +(mvx>0))>>1) + dmvector[0];
			DMV[0][1] = ((mvy  +(mvy>0))>>1) + dmvector[1] - 1;
			DMV[1][0] = ((3*mvx+(mvx>0))>>1) + dmvector[0];
			DMV[1][1] = ((3*mvy+(mvy>0))>>1) + dmvector[1] + 1;
		} else {
			DMV[0][0] = ((3*mvx+(mvx>0))>>1) + dmvector[0];
			DMV[0][1] = ((3*mvy+(mvy>0))>>1) + dmvector[1] - 1;
			DMV[1][0] = ((mvx  +(mvx>0))>>1) + dmvector[0];
			DMV[1][1] = ((mvy  +(mvy>0))>>1) + dmvector[1] + 1;
		}
	} else {
		DMV[0][0] = ((mvx+(mvx>0))>>1) + dmvector[0];
		DMV[0][1] = ((mvy+(mvy>0))>>1) + dmvector[1];
		if (picture_structure == TOP_FIELD) {
			DMV[0][1]--;
		} else {
			DMV[0][1]++;
		}
	}
}

//#define forw_frame (&(buffers[MPEG_BUFFER_FORWARD]))
//#define back_frame (&(buffers[MPEG_BUFFER_BACKWARD]))

void VDMPEGDecoder::form_predictions(int bx, int by) {
	MPEGBuffer *predframe;
	int DMV[2][2];
	int currentfield;
	int addflag = 0;

	if ((macro_block_flags & MBF_FORWARD) || (frame_type==MPEG_FRAME_TYPE_P)) {
		if (picture_structure==FRAME_PICTURE) {
			if ((motion_type==MC_FRAME) || !(macro_block_flags & MBF_FORWARD)) {
				// frame-based prediction
				form_prediction(forw_frame,0,current_frame,0,
					y_pitch,16,16,bx,by,
					PMV[0][0][0],PMV[0][0][1],0);

			} else if (motion_type==MC_FIELD) {	// field-based prediction
				// top field prediction
				form_prediction(forw_frame,mvfield_select[0][0],
					current_frame,0,y_pitch<<1,16,8,
					bx,by>>1,PMV[0][0][0],PMV[0][0][1]>>1,0);

				// bottom field prediction
				form_prediction(forw_frame,mvfield_select[1][0],
					current_frame,1,y_pitch<<1,16,8,
					bx,by>>1,PMV[1][0][0],PMV[1][0][1]>>1,0);

			} else if (motion_type==MC_DMV) {	// dual prime prediction
				// calculate derived motion vectors
				dual_prime(DMV,PMV[0][0][0],PMV[0][0][1]>>1);

				// predict top field from top field
				form_prediction(forw_frame,0,current_frame,0,
					y_pitch<<1,16,8,bx,by>>1,
					PMV[0][0][0],PMV[0][0][1]>>1,0);
				// predict and add to top field from bottom field
				form_prediction(forw_frame,1,current_frame,0,
					y_pitch<<1,16,8,bx,by>>1,
					DMV[0][0],DMV[0][1],1);

				// predict bottom field from bottom field
				form_prediction(forw_frame,1,current_frame,1,
					y_pitch<<1,16,8,bx,by>>1,
					PMV[0][0][0],PMV[0][0][1]>>1,0);
				// predict and add to bottom field from top field */
				form_prediction(forw_frame,0,current_frame,1,
					y_pitch<<1,16,8,bx,by>>1,
					DMV[1][0],DMV[1][1],1);
			}

		} else {	// TOP_FIELD or BOTTOM_FIELD

// While trying to rid myself of the "second_field" variable
// and have all buffer selection controlled externally, I've
// come to a better understanding of this next bit of code.
//
// If this is a B-field, it always predicts from the FORWARD
// buffer, even if the B-field is the second of two fields,
// because we never form predictions from B-pictures (that
// makes sense). "Second_field" is unimportant in this case.
//
// If this is a P-field, and it is the first of two P-fields,
// it always predicts from the FORWARD buffer because that's
// where the previous reference frame (or fields) reside.
//
// But if this is a P-field and it is the second of two fields
// (either I or P was the first), then there is this complication:
// If it is predicting from the field of the same parity, that
// field resides in the FORWARD buffer. If it is predicting from
// a field of opposite parity, that resides in the BACKWARD buffer.
//
// So ultimately this REQUIRES us to know whether we are decoding
// the second field of a coded frame or the first field, thus the
// "second_field" variable can't be eliminated entirely.

			// field picture
			currentfield = (picture_structure==BOTTOM_FIELD);

			// determine which frame to use for prediction
			if ((frame_type==MPEG_FRAME_TYPE_P) && second_field
				&& (currentfield!=mvfield_select[0][0])) {
				predframe = back_frame;	// same frame
			} else {
				predframe = forw_frame;	// previous frame
			}

			if ((motion_type==MC_FIELD)	|| !(macro_block_flags & MBF_FORWARD)) {
				// field-based prediction
				form_prediction(predframe,mvfield_select[0][0],
					current_frame,currentfield,y_pitch<<1,16,16,
					bx,by,PMV[0][0][0],PMV[0][0][1],0);

			} else if (motion_type==MC_16X8) {
				// predict upper half
				form_prediction(predframe,mvfield_select[0][0],
					current_frame,currentfield,y_pitch<<1,16,8,
					bx,by,PMV[0][0][0],PMV[0][0][1],0);

				// determine which frame to use for lower half prediction
				if ((frame_type==MPEG_FRAME_TYPE_P) && second_field
					&& (currentfield!=mvfield_select[1][0])) {
					predframe = back_frame;	// same frame
				} else {
					predframe = forw_frame;	// previous frame
				}

				// predict lower half
				form_prediction(predframe,mvfield_select[1][0],
					current_frame,currentfield,y_pitch<<1,16,8,
					bx,by+8,PMV[1][0][0],PMV[1][0][1],0);

			} else if (motion_type==MC_DMV) {	// dual prime prediction

// Note what the following is REALLY saying... If this is a field
// picture and it is the second of two field pictures, then it
// predicts from the previous field picture, which just happens
// to have been decoded into the same buffer we are now decoding
// into!  If it is the first of two field pictures, it predicts
// from the previous reference buffer, which contains the previous
// field (and/or frame picture) that came before.

// In short, we ALWAYS predict from the previous field/frame;
// "predframe" simply indicates in which buffer it resides.

				if (second_field) {
					predframe = back_frame;	// same frame
				} else {
					predframe = forw_frame;	// previous frame
				}

				// calculate derived motion vectors
				dual_prime(DMV,PMV[0][0][0],PMV[0][0][1]);

				// predict currentfield from field of same parity
				form_prediction(forw_frame,currentfield,current_frame,
					currentfield,y_pitch<<1,16,16,bx,by,
					PMV[0][0][0],PMV[0][0][1],0);

				// predict other field from field of opposite parity
				form_prediction(predframe,!currentfield,current_frame,
					currentfield,y_pitch<<1,16,16,bx,by,
					DMV[0][0],DMV[0][1],1);
			}
		}

		// We have fetched one prediction.  So if another
		// prediction is made then it must be bidirectional:
		addflag = 1;
	}

	if (macro_block_flags & MBF_BACKWARD) {
		if (picture_structure==FRAME_PICTURE) {
			if (motion_type==MC_FRAME) {
				// frame-based prediction
				form_prediction(back_frame,0,current_frame,0,
					y_pitch,16,16,bx,by,
					PMV[0][1][0],PMV[0][1][1],addflag);

			} else {	// field-based prediction
				// top field prediction
				form_prediction(back_frame,mvfield_select[0][1],
					current_frame,0,y_pitch<<1,16,8,
					bx,by>>1,PMV[0][1][0],PMV[0][1][1]>>1,addflag);

				// bottom field prediction
				form_prediction(back_frame,mvfield_select[1][1],
					current_frame,1,y_pitch<<1,16,8,
					bx,by>>1,PMV[1][1][0],PMV[1][1][1]>>1,addflag);
			}

		} else {	// TOP_FIELD or BOTTOM_FIELD
			// field picture
			currentfield = (picture_structure==BOTTOM_FIELD);

			if (motion_type==MC_FIELD) {
				// field-based prediction
				form_prediction(back_frame,mvfield_select[0][1],
					current_frame,currentfield,y_pitch<<1,16,16,
					bx,by,PMV[0][1][0],PMV[0][1][1],addflag);

			} else if (motion_type==MC_16X8) {
				form_prediction(back_frame,mvfield_select[0][1],
					current_frame,currentfield,y_pitch<<1,16,8,
					bx,by,PMV[0][1][0],PMV[0][1][1],addflag);

				form_prediction(back_frame,mvfield_select[1][1],
					current_frame,currentfield,y_pitch<<1,16,8,
					bx,by+8,PMV[1][1][0],PMV[1][1][1],addflag);
			}
		}
	}
}

///////////////////////////

int VDMPEGDecoder::mpeg_get_motion_component() {
	if (get_bits(1) == 0) {
		int code = show_bits(9);

		if (code >= 64) {
			code >>= 6;
			flush_buffer(MVtab0[code].len);
			code = MVtab0[code].val;
		} else if (code >= 24) {
			code >>= 3;
			flush_buffer(MVtab1[code].len);
			code = MVtab1[code].val;
		} else if (code >= 12) {
			code -= 12;
			flush_buffer(MVtab2[code].len);
			code = MVtab2[code].val;
		} else {
			SetError(kErrorBadMotionVector);
			return 0;
		}
		if (get_bits(1)) code = -code;
		return code;
	}
	return 0;
}

int VDMPEGDecoder::get_mb_address_inc() {
	int code, val;

	for (val = 0;;) {
		code = show_bits(11);
		if (code >= 24) break;
		if (code != 15) {	// if not macroblock_stuffing
			if (code == 8)	// if macroblock_escape
				val += 33;
			else {
				SetError(kErrorBadValue);
				return 1;
			}
		}
		flush_buffer(11);
	}

	if (code >= 1024) {
		flush_buffer(1);
		++val;
	} else if (code >= 128) {
		code >>= 6;
		flush_buffer(MBAtab0[code].len);
		val += MBAtab0[code].val;
	} else {
		code-= 24;
		flush_buffer(MBAtab1[code].len);
		val += MBAtab1[code].val;
	}
	return val;
}

inline int VDMPEGDecoder::get_I_mb_flags() {
// the only valid codes are 1, or 01
	int code = show_bits(2);

	if (code & 2) {
		flush_buffer(1);
		return MBF_INTRA;
	}
	if (code & 1) {
		flush_buffer(2);
	} else {
		SetError(kErrorBadValue);
	}
	return (MBF_INTRA|MBF_QUANT);
}

inline int VDMPEGDecoder::get_P_mb_flags() {
	int code = show_bits(6);

	if (code>=8) {
		code>>=3;
		flush_buffer(PMBtab0[code].len);
		code = PMBtab0[code].val;
	} else if (code) {
		flush_buffer(PMBtab1[code].len);
		code = PMBtab1[code].val;
	} else {
		SetError(kErrorBadValue);
	}
	return code;
}

inline int VDMPEGDecoder::get_B_mb_flags() {
	int code = show_bits(6);

	if (code>=8) {
		code>>=2;
		flush_buffer(BMBtab0[code].len);
		code = BMBtab0[code].val;
	} else if (code) {
		flush_buffer(BMBtab1[code].len);
		code = BMBtab1[code].val;
	} else {
		SetError(kErrorBadValue);
	}
	return code;
}

inline int VDMPEGDecoder::get_D_mb_flags() {
	if (get_bits(1)) {
		flush_buffer(1);
	} else {
		SetError(kErrorBadValue);
	}
	return MBF_INTRA;
}

inline int VDMPEGDecoder::get_mb_flags() {
	if (frame_type == MPEG_FRAME_TYPE_I) return get_I_mb_flags();
	if (frame_type == MPEG_FRAME_TYPE_P) return get_P_mb_flags();
	if (frame_type == MPEG_FRAME_TYPE_B) return get_B_mb_flags();
	if (frame_type == MPEG_FRAME_TYPE_D) return get_D_mb_flags();
	SetError(kErrorBadFrameType);
	return 0;
}

inline int VDMPEGDecoder::get_mb_cbp() {
	int code = show_bits(9);
	if (code >= 128) {
		code >>= 4;
		flush_buffer(CBPtab0[code].len);
		code = CBPtab0[code].val;
	} else if (code >= 8) {
		code >>= 1;
		flush_buffer(CBPtab1[code].len);
		code = CBPtab1[code].val;
	} else if (code) {
		flush_buffer(CBPtab2[code].len);
		code = CBPtab2[code].val;
	} else {
		SetError(kErrorBadValue);
	}
	return code;
}

// calculate motion vector component
void VDMPEGDecoder::decode_motion_vector(long& pred,
	int r_size, long code, long residual, int full_pel) {

	long lim, vec;

	lim = 16 << r_size;
	vec = full_pel? pred >> 1: pred;

	if (code > 0) {
		vec += ((code - 1) << r_size) + residual + 1;
		if (vec >= lim) vec -= lim + lim;
	} else if (code < 0) {
		vec -= ((-code - 1) << r_size) + residual + 1;
		if (vec < -lim) vec += lim + lim;
	}
	pred = full_pel? vec << 1: vec;
}

void VDMPEGDecoder::motion_vector(long *pmv,
	int h_size, int v_size, int full_pel) {

	long code, residual;

	// horizontal component (maximum h_size is 14 bits)

	code = mpeg_get_motion_component();
	residual = (h_size>0 && code!=0)? get_bits(h_size): 0;
	decode_motion_vector(pmv[0],h_size,code,residual,full_pel);

	if (motion_type==MC_DMV)
		dmvector[0] = get_bits(1)? (get_bits(1)? -1: 1): 0;

	// vertical component (maximum v_size is 14 bits)
	code = mpeg_get_motion_component();
	residual = (v_size>0 && code!=0)? get_bits(v_size): 0;
	if (mvscale) pmv[1] >>= 1;
	decode_motion_vector(pmv[1],v_size,code,residual,full_pel);
	if (mvscale) pmv[1] <<= 1;

	if (motion_type==MC_DMV)
		dmvector[1] = get_bits(1)? (get_bits(1)? -1: 1): 0;
}

void VDMPEGDecoder::motion_vectors(int s, int h_size, int v_size) {
	if (motion_vector_count == 1) {
		if (mv_format==MV_FIELD && motion_type!=MC_DMV) {
			mvfield_select[1][s] = mvfield_select[0][s] = get_bits(1);
		}
		motion_vector(PMV[0][s],h_size,v_size,0);

		PMV[1][s][0] = PMV[0][s][0];
		PMV[1][s][1] = PMV[0][s][1];
	} else {	// motion_vector_count == 2
		mvfield_select[0][s] = get_bits(1);
		motion_vector(PMV[0][s],h_size,v_size,0);

		mvfield_select[1][s] = get_bits(1);
		motion_vector(PMV[1][s],h_size,v_size,0);
	}
}

inline void VDMPEGDecoder::decode_mb(int MBA) {
	int pos_x, pos_y;
	int comp, cbp;

	macro_block_flags = get_mb_flags();
	if (mErrorState) return;

	// get frame/field motion type
	if (macro_block_flags & (MBF_FORWARD|MBF_BACKWARD)) {
		if (picture_structure==FRAME_PICTURE) {
			motion_type = frame_pred_frame_dct? MC_FRAME: get_bits(2);
		} else motion_type = get_bits(2);
	} else if ((macro_block_flags & MBF_INTRA) && concealment_vectors) {
		motion_type = (picture_structure==FRAME_PICTURE)? MC_FRAME: MC_FIELD;
	} else motion_type = 0;	// implied

	// derive motion_vector_count, mv_format
	if (picture_structure==FRAME_PICTURE) {
		motion_vector_count = (motion_type==MC_FIELD)? 2: 1;
		mv_format = (motion_type==MC_FRAME)? MV_FRAME: MV_FIELD;
	} else {
		motion_vector_count = (motion_type==MC_16X8)? 2: 1;
		mv_format = MV_FIELD;
	}

	mvscale = (mv_format==MV_FIELD && picture_structure==FRAME_PICTURE);

	// get dct_type (frame DCT / field DCT)
	dct_type = (picture_structure==FRAME_PICTURE)
		&& (!frame_pred_frame_dct)
		&& (macro_block_flags & (MBF_PATTERN|MBF_INTRA))
		? get_bits(1): 0;

	if (macro_block_flags & MBF_QUANT) {
		quantizer_scale = get_bits(5);
		if (MPEG2_Flag) {
			if (q_scale_type) {
				quantizer_scale = non_linear[quantizer_scale];
			} else quantizer_scale <<= 1;
		}
	}

	// decode forward motion vectors
	if ((macro_block_flags & MBF_FORWARD) ||
		((macro_block_flags & MBF_INTRA) && concealment_vectors)) {
		if (MPEG2_Flag)
			motion_vectors(0,f_code[0][0],f_code[0][1]);
		else
			motion_vector(PMV[0][0],f_code[0][0],
				f_code[0][0],forw_vector_full_pel);
		if (mErrorState) return;
	}

	// decode backward motion vectors
	if (macro_block_flags & MBF_BACKWARD) {
		if (MPEG2_Flag)
			motion_vectors(1,f_code[1][0],f_code[1][1]);
		else
			motion_vector(PMV[0][1],f_code[1][0],
				f_code[1][0],back_vector_full_pel);
		if (mErrorState) return;
	}

	if ((macro_block_flags & MBF_INTRA) && concealment_vectors)
		flush_buffer(1);	// remove marker_bit

	if (macro_block_flags & MBF_PATTERN) {
		cbp = get_mb_cbp();
		if (mErrorState) return;
		if (chroma_format == CHROMA422)
			cbp = (cbp << 2) | get_bits(2);
		else if (chroma_format == CHROMA444)
			cbp = (cbp << 6) | get_bits(6);
	} else
		cbp = (macro_block_flags & MBF_INTRA)?
			(1 << block_count)-1: 0;

	// decode blocks
	memset(dct_coeff, 0, sizeof(short)*64*block_count);
	cbp <<= (32-block_count);
	for (comp=0; comp<block_count; ++comp) {
		if (cbp & 0x80000000) {
			if (macro_block_flags & MBF_INTRA) {
				if (MPEG2_Flag)
					Decode_MPEG2_Intra_Block(comp);
				else
					Decode_MPEG1_Intra_Block(comp);
			} else {
				if (MPEG2_Flag)
					Decode_MPEG2_Non_Intra_Block(comp);
				else
					Decode_MPEG1_Non_Intra_Block(comp);
			}
			if (mErrorState) return;

			// IDCT
			if (cpuflags & CPUF_SUPPORTS_INTEGER_SSE)
				sse_IDCT(dct_coeff[comp]);
			else if (cpuflags & CPUF_SUPPORTS_MMX)
				mmx_IDCT(dct_coeff[comp]);
			else
				Fast_IDCT(dct_coeff[comp]);
		}
		cbp<<=1;
	}

	if (frame_type == MPEG_FRAME_TYPE_D) flush_buffer(1);

	// reset intra_dc predictors
	if (!(macro_block_flags & MBF_INTRA))
		dct_dc_y_past = dct_dc_u_past = dct_dc_v_past = 0;
	else if (!concealment_vectors) {
		PMV[0][0][0]=PMV[0][0][1]=PMV[1][0][0]=PMV[1][0][1]=0;
		PMV[0][1][0]=PMV[0][1][1]=PMV[1][1][0]=PMV[1][1][1]=0;
	}

	// special "No_MC" macroblock_type case
	if ((frame_type == MPEG_FRAME_TYPE_P)
		&& !(macro_block_flags & (MBF_FORWARD|MBF_INTRA))) {

		PMV[0][0][0]=PMV[0][0][1]=PMV[1][0][0]=PMV[1][0][1]=0;

		if (picture_structure==FRAME_PICTURE)
			motion_type = MC_FRAME;
		else {
			motion_type = MC_FIELD;
			mvfield_select[0][0] = (picture_structure==BOTTOM_FIELD);
		}
	}

	// derive current position
	pos_x = 16 * (MBA % mbWidth);
	pos_y = 16 * (MBA / mbWidth);

	cbp = ((macro_block_flags & MBF_INTRA) == 0);

	// motion compensation
	if (cbp) form_predictions(pos_x,pos_y);

	// copy or add block data into picture
	for (comp=0; comp<block_count; comp++)
		Add_Block(comp,pos_x,pos_y,cbp);
}

inline void VDMPEGDecoder::skipped_mb(int MBA) {
	int pos_x, pos_y;
	int comp;

	memset(dct_coeff, 0, sizeof(short)*64*block_count);
	// reset intra_dc predictors
	dct_dc_y_past = dct_dc_u_past = dct_dc_v_past = 0;
	// reset motion vector predictors
	if (frame_type==MPEG_FRAME_TYPE_P)
		PMV[0][0][0]=PMV[0][0][1]=PMV[1][0][0]=PMV[1][0][1]=0;

	// derive motion_type
	if (picture_structure==FRAME_PICTURE)
		motion_type = MC_FRAME;
	else {
		motion_type = MC_FIELD;
		// predict from field of same parity
		mvfield_select[0][0]=mvfield_select[0][1] =
			(picture_structure==BOTTOM_FIELD);
	}

	// skipped macroblock in I-frame is not supported
	if (frame_type == MPEG_FRAME_TYPE_I) {
		SetError(kErrorBadValue);
		return;
	}

	// derive current position
	pos_x = 16 * (MBA % mbWidth);
	pos_y = 16 * (MBA / mbWidth);

	// motion compensation
	form_predictions(pos_x,pos_y);

	// copy block data into picture
	for (comp=0; comp<block_count; comp++)
		Add_Block(comp,pos_x,pos_y,1);
}

inline void VDMPEGDecoder::slice(int pos) {
	int MBA, MBAInc, MBAMax;
	int vert_ext;

	vert_ext = (MPEG2_Flag && (pelHeight > 2800))? get_bits(3): 0;

	quantizer_scale = get_bits(5);
	if (MPEG2_Flag) {
		if (q_scale_type)
			quantizer_scale = non_linear[quantizer_scale];
		else quantizer_scale <<= 1;
	}

	while(get_bits(1)) flush_buffer(8);

	mErrorState = 0;

	MBAInc = get_mb_address_inc();
	if (mErrorState) return;
	MBA = ((vert_ext << 7) + (pos & 0xFF) - 1) * mbWidth + MBAInc - 1;
	MBAInc = 1;	// first macroblock, not skipped
	MBAMax = mbWidth * mbHeight;

	dct_dc_y_past = dct_dc_u_past = dct_dc_v_past = 0;
	PMV[0][0][0]=PMV[0][0][1]=PMV[1][0][0]=PMV[1][0][1]=0;
	PMV[0][1][0]=PMV[0][1][1]=PMV[1][1][0]=PMV[1][1][1]=0;

	while (MBA < MBAMax) {
		if (!MBAInc) {
			MBAInc = get_mb_address_inc();
			if (mErrorState) break;
		}
		if (MBAInc == 1)
			decode_mb(MBA);
		else
			skipped_mb(MBA);

		if (mErrorState) break;
		--MBAInc;
		++MBA;
	}
}

/*************************************************************

	MPEG-1 and MPEG-2 YCbCr to RGB/YUY2/UYVY conversions

*************************************************************/

// Color conversion table (see Table 6-9 of ISO/IEC 13818-2)
// Inverse coefficients are represented as 16:16 fixed point

// Confirmed output range for 8-bit R,G,B is -289 to 547
// Confirmed output range for 6-bit Green is -43 to 108
// Confirmed output range for 5-bit R,G,B is -37 to 68

#define Y_MULT	76309UL		// (255/219)*65536

static const long Inverse_Table_6_9[8][7] = {
//	crv,     cbu,     cgu,    cgv,    radd,     gadd,     badd
	{117504, 138453, -13954, -34903, -16228689, 5065481, -18910197},	// no sequence_display_extension
	{117504, 138453, -13954, -34903, -16228689, 5065481, -18910197},	// ITU-R Rec. 709 (1990)
	{104597, 132201, -25675, -53279, -14576620, 8917830, -18109946},	// unspecified
	{104597, 132201, -25675, -53279, -14576620, 8917830, -18109946},	// reserved
	{104448, 132798, -24759, -53109, -14557520, 8778945, -18186343},	// FCC
	{104597, 132201, -25675, -53279, -14576620, 8917830, -18109946},	// ITU-R Rec. 624-4 System B, G
	{104597, 132201, -25675, -53279, -14576620, 8917830, -18109946},	// SMPTE 170M
	{117579, 136230, -16907, -35559, -16238238, 5527474, -18625621}		// SMPTE 240M (1987)
};

// Confirmed range is -289 to 547, myclipper uses offset -290
static const unsigned char myclipper[840] = {
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,1,2,3,4,5,6,7,8,9,10,11,12,13,
	14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,
	30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,
	46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,
	62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,
	78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,
	94,95,96,97,98,99,100,101,102,103,104,105,106,107,108,109,
	110,111,112,113,114,115,116,117,118,119,120,121,122,123,124,125,
	126,127,128,129,130,131,132,133,134,135,136,137,138,139,140,141,
	142,143,144,145,146,147,148,149,150,151,152,153,154,155,156,157,
	158,159,160,161,162,163,164,165,166,167,168,169,170,171,172,173,
	174,175,176,177,178,179,180,181,182,183,184,185,186,187,188,189,
	190,191,192,193,194,195,196,197,198,199,200,201,202,203,204,205,
	206,207,208,209,210,211,212,213,214,215,216,217,218,219,220,221,
	222,223,224,225,226,227,228,229,230,231,232,233,234,235,236,237,
	238,239,240,241,242,243,244,245,246,247,248,249,250,251,252,253,
	254,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255
};

// Another little clip table for 15-bit conversion
// Confirmed range is -37 to 68, myclipper2 uses offset -38
static const unsigned char myclipper2[108] = {
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,1,2,3,4,5,6,7,8,9,
	10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
	26,27,28,29,30,31,31,31,31,31,31,31,31,31,31,31,
	31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,
	31,31,31,31,31,31,31,31,31,31,31,31
};

// Another little clip table for 6-bit green
// Confirmed range is -43 to 108, myclipper3 uses offset -44
static const unsigned char myclipper3[156] = {
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,1,2,3,
	4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,
	20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,
	36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,
	52,53,54,55,56,57,58,59,60,61,62,63,63,63,63,63,
	63,63,63,63,63,63,63,63,63,63,63,63,63,63,63,63,
	63,63,63,63,63,63,63,63,63,63,63,63,63,63,63,63,
	63,63,63,63,63,63,63,63,63,63,63,63
};



void VDMPEGDecoder::EnableMatrixCoefficients(bool bEnable) {
	fAllowMatrixCoefficients = bEnable;
}

// Upsample 4:2:0 chroma to 4:2:2
// (See ISO/IEC 13818-2 section 6.1.2.1)

void VDMPEGDecoder::mpeg_conv420to422(YCCSample *src,
	YCCSample *dst, long w, long h, bool progressive) {

	YCCSample *sm1, *sp1;
	int x,y;

	if (progressive) {
		// MPEG-1 or MPEG-2 progressive frame

		h -= 1;
		for (y = 0; y <= h; ++y) {
			sm1 = (y > 0)? src-w: src;
			sp1 = (y < h)? src+w: src;
			for (x = 0; x < w; ++x) {
				YCCSample s = *src++;
				dst[w] = (YCCSample)((s + *sp1++ + 1) >> 1);
				*dst++ = (YCCSample)((s + *sm1++ + 1) >> 1);
			}
			dst += w;
		}

	} else {
		// MPEG-2 interlaced frame
		int ym1, y1, yp1;

		h = (h >> 1) - 1;

		for (y = 0; y <= h; ++y) {
			// top field
			sm1 = (y > 0)? src-w*2: src;
			sp1 = (y < h)? src+w*2: src;
			for (x = 0; x < w; ++x) {
				ym1 = *sm1++;
				y1  = *src++;
				yp1 = *sp1++;
				dst[w*2] = (YCCSample)((y1*5 + yp1*3 + 4) >> 3);
				*dst++ = (YCCSample)((y1*7 + ym1 + 4) >> 3);
			}

			// bottom field
			sm1 = (y > 0)? src-w*2: src;
			sp1 = (y < h)? src+w*2: src;
			for (x = 0; x < w; ++x) {
				ym1 = *sm1++;
				y1  = *src++;
				yp1 = *sp1++;
				dst[w*2] = (YCCSample)((y1*7 + yp1 + 4) >> 3);
				*dst++ = (YCCSample)((y1*5 + ym1*3 + 4) >> 3);
			}

			dst += w*2;
		}
	}
}

// Upsample 4:2:2 chroma to 4:4:4
// (See ISO/IEC 13818-2 section 6.1.2.2)

void VDMPEGDecoder::mpeg_conv422to444(YCCSample *src,
	YCCSample *dst, long w, long h, int mpeg2) {

	int xm1, x1, xp1;
	int x,y;

	if (!mpeg2) {

		for (y = 0; y < h; ++y) {
			x1 = xp1 = *src++;
			for (x = 1; x < w; ++x) {
				xm1 = x1;
				x1 = xp1;
				xp1 = *src++;
				*dst++ = (YCCSample)((xm1 + x1 + 1) >> 1);
				*dst++ = (YCCSample)((x1 + xp1 + 1) >> 1);
			}
			*dst++ = (YCCSample)((x1 + xp1 + 1) >> 1);
			*dst++ = xp1;
		}

	} else {

		for (y = 0; y < h; ++y) {
			xp1 = *src++;
			for (x = 1; x < w; ++x) {
				x1 = xp1;
				xp1 = *src++;
				*dst++ = (YCCSample)x1;
				*dst++ = (YCCSample)((x1 + xp1 + 1) >> 1);
			}
			*dst++ = (YCCSample)xp1;
			*dst++ = (YCCSample)xp1;
		}
	}
}


// Convert YCbCr 4:4:4 to RGB 16,24,32
// (See ISO/IEC 13818-2 section 6.3.6)

inline void VDMPEGDecoder::mpeg_conv444toRGB(
	YCCSample *srcY, YCCSample *srcU, YCCSample *srcV,
	void *_dst, int bpr, int w, int h, int bpp) {

// bpr is the pitch, and it can be negative!
// If so, dst points at the END of the bitmap

    const long crv = Inverse_Table_6_9[matrix_coefficients&7][0];
    const long cbu = Inverse_Table_6_9[matrix_coefficients&7][1];
    const long cgu = Inverse_Table_6_9[matrix_coefficients&7][2];
    const long cgv = Inverse_Table_6_9[matrix_coefficients&7][3];
    const long radd = Inverse_Table_6_9[matrix_coefficients&7][4];
    const long gadd = Inverse_Table_6_9[matrix_coefficients&7][5];
    const long badd = Inverse_Table_6_9[matrix_coefficients&7][6];

    unsigned char const *clipper = myclipper+290;
	unsigned char const *clipper2 = myclipper2+38;
	unsigned char const *clipper3 = myclipper3+44;

    long ht,y,u,v;
	int x;

	switch (bpp) {

		case 15: {
			unsigned short *dst;
			unsigned short i;

			for (ht = 0; ht < h; ++ht) {
				dst = (unsigned short *)((char *)_dst + ht * bpr);

				for (x = 0; x < w; ++x) {
					y = srcY[x] * Y_MULT;
					u = srcU[x];
					v = srcV[x];

					i = clipper2[(y + v*crv + radd) >> 19];
					i = (i << 5) + clipper2[(y + u*cgu + v*cgv + gadd) >> 19];
					*dst++ = (i << 5) + clipper2[(y + u*cbu + badd) >> 19];
				}

				srcY += y_pitch;
				srcU += y_pitch;
				srcV += y_pitch;
			}
		}
		break;

		case 16: {
			unsigned short *dst;
			unsigned short i;

			for (ht = 0; ht < h; ++ht) {
				dst = (unsigned short *)((char *)_dst + ht * bpr);

				for (x = 0; x < w; ++x) {
					y = srcY[x] * Y_MULT;
					u = srcU[x];
					v = srcV[x];

					i = clipper2[(y + v*crv + radd) >> 19];
					i = (i << 6) + clipper3[(y + u*cgu + v*cgv + gadd) >> 18];
					*dst++ = (i << 5) + clipper2[(y + u*cbu + badd) >> 19];
				}

				srcY += y_pitch;
				srcU += y_pitch;
				srcV += y_pitch;
			}
		}
		break;

		case 24: {
			unsigned char *dst;

			for (ht = 0; ht < h; ++ht) {
				dst = (unsigned char *)_dst + ht * bpr;

				for (x = 0; x < w; ++x) {
					y = srcY[x] * Y_MULT;
					u = srcU[x];
					v = srcV[x];

					*dst++ = clipper[(y + u*cbu + badd) >> 16];
					*dst++ = clipper[(y + u*cgu + v*cgv + gadd) >> 16];
					*dst++ = clipper[(y + v*crv + radd) >> 16];
				}

				srcY += y_pitch;
				srcU += y_pitch;
				srcV += y_pitch;
			}
		}
		break;

		case 32: {
			unsigned long *dst;
			unsigned long i;

			for (ht = 0; ht < h; ++ht) {
				dst = (unsigned long *)((char *)_dst + ht * bpr);

				for (x = 0; x < w; ++x) {
					y = srcY[x] * Y_MULT;
					u = srcU[x];
					v = srcV[x];

					i = clipper[(y + v*crv + radd) >> 16];
					i = (i << 8) + clipper[(y + u*cgu + v*cgv + gadd) >> 16];
					*dst++ = (i << 8) + clipper[(y + u*cbu + badd) >> 16];
				}

				srcY += y_pitch;
				srcU += y_pitch;
				srcV += y_pitch;
			}
		}
		break;
	}
}

// Convert YCbCr to RGB 16,24,32

void VDMPEGDecoder::mpeg_YUVToRGB(int buffer_ID,
	void *dst, long bpr, long w, long h, int bpp) {

	YCCSample *Y_ptr = buffers[buffer_ID].Y;
	YCCSample *U_ptr = buffers[buffer_ID].U;
	YCCSample *V_ptr = buffers[buffer_ID].V;

	YCCSample *u444ptr = u444;
	YCCSample *v444ptr = v444;

	// Coded_Picture_Width, Coded_Picture_Height
	long cpw = (pelWidth + 15) & -16;
	long cph = (pelHeight + 15) & -16;

	// Progressive or Interlaced?
	const bool pf = !!(buffers[buffer_ID].bitflags & PICTUREFLAG_PF);

	switch (chroma_format) {

	case CHROMA420:
		if (cpuflags & CPUF_SUPPORTS_MMX) {
			if (cpuflags & CPUF_SUPPORTS_INTEGER_SSE) {
				isse_conv420to422(U_ptr, uv422, cpw >> 1, cph >> 1, pf);
				isse_conv422to444(uv422, u444ptr, cpw >> 1, cph, MPEG2_Flag);
				isse_conv420to422(V_ptr, uv422, cpw >> 1, cph >> 1, pf);
				isse_conv422to444(uv422, v444ptr, cpw >> 1, cph, MPEG2_Flag);
				__asm sfence
			} else {
				mmx_conv420to422(U_ptr, uv422, cpw >> 1, cph >> 1, pf);
				mmx_conv422to444(uv422, u444ptr, cpw >> 1, cph, MPEG2_Flag);
				mmx_conv420to422(V_ptr, uv422, cpw >> 1, cph >> 1, pf);
				mmx_conv422to444(uv422, v444ptr, cpw >> 1, cph, MPEG2_Flag);
			}
			__asm emms
		} else {
			mpeg_conv420to422(U_ptr, uv422, cpw >> 1, cph >> 1, pf);
			mpeg_conv422to444(uv422, u444ptr, cpw >> 1, cph, MPEG2_Flag);
			mpeg_conv420to422(V_ptr, uv422, cpw >> 1, cph >> 1, pf);
			mpeg_conv422to444(uv422, v444ptr, cpw >> 1, cph, MPEG2_Flag);
		}
		break;

	case CHROMA422:
		if (cpuflags & CPUF_SUPPORTS_MMX) {
			if (cpuflags & CPUF_SUPPORTS_INTEGER_SSE) {
				isse_conv422to444(U_ptr, u444ptr, cpw >> 1, cph, MPEG2_Flag);
				isse_conv422to444(V_ptr, v444ptr, cpw >> 1, cph, MPEG2_Flag);
				__asm sfence
			} else {
				mmx_conv422to444(U_ptr, u444ptr, cpw >> 1, cph, MPEG2_Flag);
				mmx_conv422to444(V_ptr, v444ptr, cpw >> 1, cph, MPEG2_Flag);
			}
			__asm emms
		} else {
			mpeg_conv422to444(U_ptr, u444ptr, cpw >> 1, cph, MPEG2_Flag);
			mpeg_conv422to444(V_ptr, v444ptr, cpw >> 1, cph, MPEG2_Flag);
		}
		break;

	case CHROMA444:
		u444ptr = U_ptr;
		v444ptr = V_ptr;
		break;
	}

	mpeg_conv444toRGB(Y_ptr, u444ptr, v444ptr, dst, bpr, w, h, bpp);
}

// Convert YCbCr to either YUYV or YVYU
bool VDMPEGDecoder::mpeg_YUVToYUY2(void *dst, ptrdiff_t pitch, int buffer,
	YCCSample *Y_ptr, YCCSample *U_ptr, YCCSample *V_ptr) {

	if (memblock != NULL && dst != NULL && pitch > 0 && (unsigned int)buffer < MPEG_BUFFERS) {

		YCCSample *u422ptr = u444;
		YCCSample *v422ptr = v444;

		// Coded_Picture_Width, Coded_Picture_Height
		long cpw = (pelWidth + 15) & -16;
		long cph = (pelHeight + 15) & -16;

		// Progressive or Interlaced?
		const bool pf = !!(buffers[buffer].bitflags & PICTUREFLAG_PF);

		switch (chroma_format) {

		case CHROMA420:
			if (cpuflags & CPUF_SUPPORTS_MMX) {
				if (cpuflags & CPUF_SUPPORTS_INTEGER_SSE) {
					isse_conv420to422(U_ptr, u422ptr, cpw >> 1, cph >> 1, pf);
					isse_conv420to422(V_ptr, v422ptr, cpw >> 1, cph >> 1, pf);
					__asm sfence
				} else {
					mmx_conv420to422(U_ptr, u422ptr, cpw >> 1, cph >> 1, pf);
					mmx_conv420to422(V_ptr, v422ptr, cpw >> 1, cph >> 1, pf);
				}
				mmx_conv422toYUY2(Y_ptr, u422ptr, v422ptr, dst, cpw >> 1, cph);
				__asm emms
				return true;	// Done
			}
			mpeg_conv420to422(U_ptr, u422ptr, cpw >> 1, cph >> 1, pf);
			mpeg_conv420to422(V_ptr, v422ptr, cpw >> 1, cph >> 1, pf);
			break;

		case CHROMA422:
			if (cpuflags & CPUF_SUPPORTS_MMX) {
				mmx_conv422toYUY2(Y_ptr, U_ptr, V_ptr, dst, cpw >> 1, cph);
				__asm emms
				return true;	// Done
			}
			u422ptr = U_ptr;
			v422ptr = V_ptr;
			break;

		case CHROMA444:
			// Currently unsupported
			return false;
		}

		YCCSample *dest = (YCCSample *)dst;
		for (long x = (cpw >> 1) * cph; x > 0; --x) {
			*dest++ = *Y_ptr++;
			*dest++ = *u422ptr++;
			*dest++ = *Y_ptr++;
			*dest++ = *v422ptr++;
		}
		return true;
	}

	return false;
}

///////////////////////////////////////////////////////
//
//		Public
//
///////////////////////////////////////////////////////

const void *VDMPEGDecoder::GetYBuffer(int buffer, ptrdiff_t& pitch) {
	if ((unsigned int)buffer < MPEG_BUFFERS) {
		pitch = y_pitch;
		return buffers[buffer].Y;
	}
	return NULL;
}

const void *VDMPEGDecoder::GetCbBuffer(int buffer, ptrdiff_t& pitch) {
	if ((unsigned int)buffer < MPEG_BUFFERS) {
		pitch = (chroma_format == CHROMA420)? y_pitch >> 1: y_pitch;
		return buffers[buffer].U;
	}
	return NULL;
}

const void *VDMPEGDecoder::GetCrBuffer(int buffer, ptrdiff_t& pitch) {
	if ((unsigned int)buffer < MPEG_BUFFERS) {
		pitch = (chroma_format == CHROMA420)? y_pitch >> 1: y_pitch;
		return buffers[buffer].V;
	}
	return NULL;
}

// Convert YCbCr to UYVY
bool VDMPEGDecoder::DecodeUYVY(void *dst, ptrdiff_t pitch, int buffer) {

	if (memblock != NULL && dst != NULL && pitch > 0 && (unsigned int)buffer < MPEG_BUFFERS) {

		YCCSample *Y_ptr = buffers[buffer].Y;
		YCCSample *U_ptr = buffers[buffer].U;
		YCCSample *V_ptr = buffers[buffer].V;

		YCCSample *u422ptr = u444;
		YCCSample *v422ptr = v444;

		// Coded_Picture_Width, Coded_Picture_Height
		long cpw = (pelWidth + 15) & -16;
		long cph = (pelHeight + 15) & -16;

		// Progressive or Interlaced?
		const bool pf = !!(buffers[buffer].bitflags & PICTUREFLAG_PF);

		switch (chroma_format) {

			case CHROMA420:
				if (cpuflags & CPUF_SUPPORTS_MMX) {
					if (cpuflags & CPUF_SUPPORTS_INTEGER_SSE) {
						mmx_conv420to422(U_ptr, u422ptr, cpw >> 1, cph >> 1, pf);
						mmx_conv420to422(V_ptr, v422ptr, cpw >> 1, cph >> 1, pf);
						__asm sfence
					} else {
						mmx_conv420to422(U_ptr, u422ptr, cpw >> 1, cph >> 1, pf);
						mmx_conv420to422(V_ptr, v422ptr, cpw >> 1, cph >> 1, pf);
					}
					mmx_conv422toUYVY(Y_ptr, u422ptr, v422ptr, dst, cpw >> 1, cph);
					__asm emms
					return true;	// Done
				}
				mpeg_conv420to422(U_ptr, u422ptr, cpw >> 1, cph >> 1, pf);
				mpeg_conv420to422(V_ptr, v422ptr, cpw >> 1, cph >> 1, pf);
				break;

			case CHROMA422:
				if (cpuflags & CPUF_SUPPORTS_MMX) {
					mmx_conv422toUYVY(Y_ptr, U_ptr, V_ptr, dst, cpw >> 1, cph);
					__asm emms
					return true;	// Done
				}
				u422ptr = U_ptr;
				v422ptr = V_ptr;
				break;

			case CHROMA444:
				// Currently unsupported
				return false;
		}

		YCCSample *dest = (YCCSample *)dst;
		for (long x = (cpw >> 1) * cph; x > 0; --x) {
			*dest++ = *u422ptr++;
			*dest++ = *Y_ptr++;
			*dest++ = *v422ptr++;
			*dest++ = *Y_ptr++;
		}
		return true;
	}

	return false;
}

// Convert YCbCr to YUY2
bool VDMPEGDecoder::DecodeYUYV(void *dst, ptrdiff_t pitch, int buffer) {
	if (memblock != NULL && dst != NULL && pitch && (unsigned int)buffer < MPEG_BUFFERS) {
		return mpeg_YUVToYUY2(dst, pitch, buffer,
			buffers[buffer].Y, buffers[buffer].U, buffers[buffer].V);
	}
	return false;
}

// Convert YCbCr to YVYU
bool VDMPEGDecoder::DecodeYVYU(void *dst, ptrdiff_t pitch, int buffer) {
	if (memblock != NULL && dst != NULL && pitch && (unsigned int)buffer < MPEG_BUFFERS) {
		return mpeg_YUVToYUY2(dst, pitch, buffer,
			buffers[buffer].Y, buffers[buffer].V, buffers[buffer].U);
	}
	return false;
}

// Y41P (untested)
bool VDMPEGDecoder::DecodeY41P(void *dst, ptrdiff_t pitch, int buffer) {

	if (memblock != NULL && dst != NULL && pitch > 0 && (unsigned int)buffer < MPEG_BUFFERS) {

		YCCSample *Y_ptr = buffers[buffer].Y;
		YCCSample *U_ptr = buffers[buffer].U;
		YCCSample *V_ptr = buffers[buffer].V;

		YCCSample *u422ptr = u444;
		YCCSample *v422ptr = v444;

		// Coded_Picture_Width, Coded_Picture_Height
		long cpw = (pelWidth + 15) & -16;
		long cph = (pelHeight + 15) & -16;

		long x, y;

		// Progressive or Interlaced?
		const bool pf = !!(buffers[buffer].bitflags & PICTUREFLAG_PF);

		// Start by converting 4:2:0 to 4:2:2
		switch (chroma_format) {

			case CHROMA420:
				if (cpuflags & CPUF_SUPPORTS_MMX) {
					if (cpuflags & CPUF_SUPPORTS_INTEGER_SSE) {
						isse_conv420to422(U_ptr, u422ptr, cpw >> 1, cph >> 1, pf);
						isse_conv420to422(V_ptr, v422ptr, cpw >> 1, cph >> 1, pf);
						__asm sfence
					} else {
						mmx_conv420to422(U_ptr, u422ptr, cpw >> 1, cph >> 1, pf);
						mmx_conv420to422(V_ptr, v422ptr, cpw >> 1, cph >> 1, pf);
					}
					__asm emms
				} else {
					mpeg_conv420to422(U_ptr, u422ptr, cpw >> 1, cph >> 1, pf);
					mpeg_conv420to422(V_ptr, v422ptr, cpw >> 1, cph >> 1, pf);
				}
				break;

			case CHROMA422:
				u422ptr = U_ptr;
				v422ptr = V_ptr;
				break;

			case CHROMA444:
				// Currently unsupported
				return false;
		}

// Y41P format uses 4:1 horizontal downsampling and no vertical
// downsampling. Y is sampled at every pixel; U and V are sampled
// every 4 pixels horizontally. Each macropixel is 12 bytes and
// contains 8 pixels. It uses the following byte order:

// U0 Y0 V0 Y1   U4 Y2 V4 Y3   Y4 Y5 Y6 Y7

		for (y = 0; y < cph; ++y) {
			YCCSample *dest = (YCCSample *)((char *)dst + y * pitch);

			for (x = (cpw >> 3); x > 0; x--) {
				*dest++ = *Y_ptr++;
				*dest++ = *v422ptr++;
				*dest++ = *Y_ptr++;
				*dest++ = *u422ptr++;
				v422ptr++;
				u422ptr++;

				*dest++ = *Y_ptr++;
				*dest++ = *v422ptr++;
				*dest++ = *Y_ptr++;
				*dest++ = *u422ptr++;
				v422ptr++;
				u422ptr++;

				*((long *)dest) = *((long *)Y_ptr);
				dest += 4;
				Y_ptr += 4;
			}
		}

		return true;
	}

	return false;
}

bool VDMPEGDecoder::DecodeRGB15(void *dst, ptrdiff_t pitch, int buffer) {
	if (memblock != NULL && dst != NULL && pitch && (unsigned int)buffer < MPEG_BUFFERS) {
		mpeg_YUVToRGB(buffer, dst, pitch, pelWidth, pelHeight, 15);
		return true;
    }
	return false;
}

// mpDecoder->DecodeRGB16(pBuffer + w*2*(h-1), -w*2, buffer);
bool VDMPEGDecoder::DecodeRGB16(void *dst, ptrdiff_t pitch, int buffer) {
	if (memblock != NULL && dst != NULL && pitch && (unsigned int)buffer < MPEG_BUFFERS) {
		mpeg_YUVToRGB(buffer, dst, pitch, pelWidth, pelHeight, 16);
		return true;
    }
	return false;
}

// mpDecoder->DecodeRGB24(pBuffer + w*3*(h-1), -w*3, buffer);
bool VDMPEGDecoder::DecodeRGB24(void *dst, ptrdiff_t pitch, int buffer) {
	if (memblock != NULL && dst != NULL && pitch && (unsigned int)buffer < MPEG_BUFFERS) {
		mpeg_YUVToRGB(buffer, dst, pitch, pelWidth, pelHeight, 24);
		return true;
    }
	return false;
}

// mpDecoder->DecodeRGB32(pBuffer + w*4*(h-1), -w*4, buffer);
bool VDMPEGDecoder::DecodeRGB32(void *dst, ptrdiff_t pitch, int buffer) {
	if (memblock != NULL && dst != NULL && pitch && (unsigned int)buffer < MPEG_BUFFERS) {
		mpeg_YUVToRGB(buffer, dst, pitch, pelWidth, pelHeight, 32);
		return true;
    }
	return false;
}
