//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2005 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//	This version has been modified by fccHandler, July 2006

#include <stdafx.h>
#include <crtdbg.h>
#include <math.h>

#include <windows.h>

#include "InputFileWMV.h"
#include "AudioSource.h"
#include "VideoSource.h"
#include "WMVTypes.h"
#include "ProgressDialog.h"
#include <vd2/system/error.h>
#include <vd2/system/refcount.h>
#include <vd2/Dita/resources.h>

#include "gui.h"
#include "DataVector.h"

#define PATCH_WMA
//#define DISABLE_ASF
#define ASF_VBR_FIX

// HACK!!!!

extern CRITICAL_SECTION g_diskcs;
extern bool g_disklockinited;

namespace {
	enum { kVDST_InputFileASF = 11 };

	enum {
		kVDM_OpeningFile,			// WMV: Opening file "%ls"
		kVDM_DataObjectTruncated,   // WMV: File is incomplete, some video/audio data may be missing.
	};
}

struct ASFSampleInfo {
	__int64			pos;
	long			size;
	long			pres_time;
	unsigned char	obj_number;
	unsigned char	flags;	// key frame bit
};

// Flags for ASFSampleInfo.flags
#define ASI_FLAG_KEYFRAME				1
#define ASI_FLAG_SEEN_FIRST_FRAGMENT	2

///////////////////////////////////////////////////////////////////////////

// Get rid of the useless "ASFIndex" class and
// reuse the "DataVector" class ripped from Mpeg.cpp

class ASFStreamNode : public ListNode2<ASFStreamNode> {
public:
	int						stream_number;
	AVISTREAMINFO			hdr;
	void					*pFormat;
	long					lFormatLen;
//	DataVector				dv;
	DataVector				*dv;
	ASFSampleInfo			*sample_list;
	__int64					bytes;
	long					lLength;
	int						iSpan;
	long					lAudioScrambleChunk;

	ASFStreamNode();
	~ASFStreamNode();
};

// Maybe better (memory-wise) to manage "dv" explicitly

//ASFStreamNode::ASFStreamNode() : dv(sizeof ASFSampleInfo) {
ASFStreamNode::ASFStreamNode() {

	pFormat = NULL;
	bytes = 0;

	iSpan = 1;		// The value 1 has no effect

	sample_list = NULL;
	dv = NULL;
}

ASFStreamNode::~ASFStreamNode() {
	if (sample_list)	delete [] sample_list;
	if (pFormat)		delete [] pFormat;
	if (dv)				delete dv;
}

///////////////////////////////////////////////////////////////////////////

class ASFReadHandler : public IAVIReadHandler {
public:
	bool		fDisableFastIO;
	bool		fQuiet;
	bool		fAcceptPartial;

	VDFraction mFrameRate;
	long		lPreroll;

	ASFReadHandler(const wchar_t *, bool, bool);
	~ASFReadHandler();

	void AddRef();
	void Release();
	IAVIReadStream *GetStream(DWORD fccType, LONG lParam);
	void EnableFastIO(bool);
	bool isOptimizedForRealtime();
	bool isStreaming();
	bool isIndexFabricated();
	bool AppendFile(const wchar_t *pszFile);
	bool getSegmentHint(const char **ppszPath);

	void GetTextInfo(tTextInfo& textInfo);
	void GetTextInfoEncoding(int& codePage, int& countryCode, int& language, int& dialect);

	long ReadData(int streamno, void *lpBuffer, __int64 pktpos, int offset, long bytes, ASFSampleInfo *asi);

private:
	int ref_count;

// Sheesh, buffer the file input please
	enum { READ_BUFFER_SIZE = 16384 };
	unsigned char *read_buffer;
	long _refill_buffer();

	HANDLE hFile;
	__int64		i64FilePosition;
	__int64		i64MovieLimit;
	__int64		i64Size;
	long		pktsize;
	BYTE		*pktbuf;

	int			mTextInfoCodePage;
	int			mTextInfoCountryCode;
	int			mTextInfoLanguage;
	int			mTextInfoDialect;

	tTextInfo	mTextInfo;

	List2<ASFStreamNode>	listStreams;

	void		_construct(const wchar_t *pszFile);
	void		_parseFileRoot();
	void		_parse_ASF_Header_Object(__int64 len);
	void		_parse_ASF_Header_Extension_Object(__int64 len);
	void		_parse_ASF_Extended_Stream_Properties_Object(__int64 len);
	void		_parse_ASF_File_Properties_Object(__int64 len);
	void		_parse_ASF_Content_Description_Object(__int64 len);
	void		_parse_ASF_Stream_Properties_Object(__int64 len);
	int			_get_VLC(const void *data, DWORD *VLC, int flags);
	void		_parse_ASF_Data_Object(__int64 pos, __int64 len);
	void		_destruct();

	long		_readFile(void *data, long len);
	void		_seekFile(__int64 i64NewPos);
	__int64		_posFile();
};

IAVIReadHandler *CreateASFReadHandler(const wchar_t *s, bool fQuiet, bool fAcceptPartial) {
	return new ASFReadHandler(s, fQuiet, fAcceptPartial);
}

///////////////////////////////////////////////////////////////////////////

class ASFReadStream : public IAVIReadStream {
	friend ASFReadHandler;

public:
	ASFReadStream(ASFReadHandler *, ASFStreamNode *);
	~ASFReadStream();

	HRESULT BeginStreaming(long lStart, long lEnd, long lRate);
	HRESULT EndStreaming();
	HRESULT Info(AVISTREAMINFO *pasi, long lSize);
	bool IsKeyFrame(long lFrame);
	HRESULT Read(VDPosition lStart64, long lSamples, void *lpBuffer, long cbBuffer, long *plBytes, long *plSamples);
	VDPosition Start();
	VDPosition End();
	long PrevKeyFrame(long lFrame);
	long NextKeyFrame(long lFrame);
	long NearestKeyFrame(long lFrame);
	HRESULT FormatSize(long lFrame, long *plSize);
	HRESULT ReadFormat(long lFrame, void *pFormat, long *plSize);
	bool isStreaming();
	bool isKeyframeOnly();

	bool getVBRInfo(double& bitrate_mean, double& bitrate_stddev, double& maxdev);
	sint64 getSampleBytePosition(VDPosition sample_num);

	VDPosition	TimeToPosition(VDTime timeInMicroseconds);
	VDTime		PositionToTime(VDPosition pos);

private:
	void UnscrambleAudio(long& lStart, long& lSamples);
#ifdef ASF_VBR_FIX
	long TimeToSample(long lTimeInMs);
#endif

	ASFReadHandler *parent;
	ASFStreamNode *psnData;
	ASFSampleInfo *pIndex;
	long length;
	int Stream_Number;

	__int64			i64Pos;
	long			lCurrent;
};

///////////////////////////////////////////////////////////////////////////

ASFReadStream::ASFReadStream(ASFReadHandler *parent, ASFStreamNode *psnData) {
	this->parent = parent;
	this->psnData = psnData;
	this->Stream_Number = psnData->stream_number;

	parent->AddRef();

	pIndex = psnData->sample_list;
	length = psnData->lLength;

	i64Pos = 0;
	lCurrent = 0;
}

ASFReadStream::~ASFReadStream() {
	parent->Release();
}

HRESULT ASFReadStream::BeginStreaming(long lStart, long lEnd, long lRate) {
	return AVIERR_OK;
}

HRESULT ASFReadStream::EndStreaming() {
	return AVIERR_OK;
}

HRESULT ASFReadStream::Info(AVISTREAMINFO *pasi, long lSize) {
/*
	AVISTREAMINFO asi;

	memset(&asi, 0, sizeof asi);

	asi.fccType					= psnData->hdr.fccType;
	asi.fccHandler				= psnData->hdr.fccHandler;
	asi.dwFlags					= psnData->hdr.dwFlags;
	asi.wPriority				= psnData->hdr.wPriority;
	asi.wLanguage				= psnData->hdr.wLanguage;
	asi.dwScale					= psnData->hdr.dwScale;
	asi.dwRate					= psnData->hdr.dwRate;
	asi.dwStart					= psnData->hdr.dwStart;
	asi.dwLength				= psnData->hdr.dwLength;
	asi.dwInitialFrames			= psnData->hdr.dwInitialFrames;
	asi.dwSuggestedBufferSize	= psnData->hdr.dwSuggestedBufferSize;
	asi.dwQuality				= psnData->hdr.dwQuality;
	asi.dwSampleSize			= psnData->hdr.dwSampleSize;
	asi.rcFrame.top				= psnData->hdr.rcFrame.top;
	asi.rcFrame.left			= psnData->hdr.rcFrame.left;
	asi.rcFrame.right			= psnData->hdr.rcFrame.right;
	asi.rcFrame.bottom			= psnData->hdr.rcFrame.bottom;

	if (lSize < sizeof asi) {
		memcpy(pasi, &asi, lSize);
	} else {
		memcpy(pasi, &asi, sizeof asi);
		memset((char *)pasi + sizeof asi, 0, lSize - sizeof asi);
	}
*/

	// KISS...

	const int hdrsize = sizeof(psnData->hdr); //AVISTREAMINFO);

	if (lSize <= hdrsize) {
		memcpy(pasi, &psnData->hdr, lSize);
	} else {
		memcpy(pasi, &psnData->hdr, hdrsize);
		memset((char *)pasi + hdrsize, 0, lSize - hdrsize);
	}

	return AVIERR_OK;
}

bool ASFReadStream::IsKeyFrame(long lFrame) {
	if (lFrame >= 0 && lFrame < length) {
		if (psnData->hdr.fccType == streamtypeAUDIO)
			return true;

		return (pIndex[lFrame].flags & ASI_FLAG_KEYFRAME);
	}
	return false;
}


void ASFReadStream::UnscrambleAudio(long& lStart, long& lSamples) {

	// Check for repeating.

	if (lStart < 0) {
		lStart = 0;
		lSamples = 1;
	}

	// Check for audio scrambling.

	if (psnData->iSpan > 1) {

		// Audio is scrambled (sh*t!).
		//
		// ASF1 seems to scramble data differently than described in the
		// ASF2 specs; it's a simple index reversal.  Assuming each
		// scrambling block is composed of sample[a][b], where
		// b=0...(scramble_chunk_size-1), the unscrambling method is
		// simply sample[b][a] (with index dimensions reversed as well,
		// of course).
		//
		// For now, we'll be lazy, and simply decode one sample at a time,
		// reducing the read to a single index transform.  This REALLY
		// should be changed!

		long scramble_base;
		long scramble_offset;
		int i, j;

		// Break down sample number into (start of scrambled block) and
		// (offset within block).

		scramble_offset = lStart % (psnData->iSpan * psnData->lAudioScrambleChunk);
		scramble_base = lStart - scramble_offset;

		// Check to see if this is a partial block and abort descramble if so

		if (scramble_base + psnData->iSpan * psnData->lAudioScrambleChunk <= psnData->lLength) {

			// Descramble offset within block.

			i = scramble_offset / psnData->iSpan;
			j = scramble_offset % psnData->iSpan;

			scramble_offset = i + j * psnData->lAudioScrambleChunk;

			// Combine to new sample number.

			lStart = scramble_base + scramble_offset;
			lSamples = 1;
		}

		_RPT3(0,"new lStart = %d, length = %d, count = %d\n", lStart, psnData->lLength, lSamples);

//		This code works for NGE_01.ASF with SVAL=3, SVAL2=21
//		lSamples = 1;
//		lStart = ((lStart / SVAL2) % SVAL) + (lStart%SVAL2)*SVAL + (lStart/(SVAL*SVAL2))*SVAL*SVAL2;
	}
}


// FIXME: The following doesn't work with VBR audio...

// Interesting contrast between WMA VBR and MP3 VBR:
// MP3 -- blocks are all the same duration, but different sizes
// WMA -- blocks are all the same size, but different durations

HRESULT ASFReadStream::Read(VDPosition lStart64, long lSamples, void *lpBuffer, long cbBuffer, long *plBytes, long *plSamples) {
	long lActual;
	long lStart = (long)lStart64;

	if (lStart < 0 || lStart >= length || (lSamples <= 0 && lSamples != AVISTREAMREAD_CONVENIENT)) {
		// umm... dummy!  can't read outside of stream!

		if (plBytes) *plBytes = 0;
		if (plSamples) *plSamples = 0;

		return AVIERR_OK;
	}

	// blocked or discrete?

	if (psnData->hdr.fccType == streamtypeAUDIO) {
		long sampsize = psnData->hdr.dwSampleSize;

		// Handle repeating and/or scrambled audio
		UnscrambleAudio(lStart, lSamples);

		__int64 byte_off = (__int64)lStart * sampsize;
		long bytecnt;

		// too small to hold a sample?

		if (lpBuffer && cbBuffer < sampsize)
			return AVIERR_BUFFERTOOSMALL;

		if (byte_off >= i64Pos) {
			// Start from current position
			byte_off -= i64Pos;
		} else {
			// Start from the beginning
			i64Pos = 0;
			lCurrent = 0;
		}

		while (byte_off >= (pIndex[lCurrent].size)) {
			byte_off -= (pIndex[lCurrent].size);
			i64Pos += (pIndex[lCurrent].size);
			lCurrent++;
		}

		// Client too lazy to specify a size?

		if (lSamples == AVISTREAMREAD_CONVENIENT) {
			lSamples = (pIndex[lCurrent].size - (long)byte_off) / sampsize;

			if (!lSamples && (lCurrent+1 < length))
				lSamples = (pIndex[lCurrent].size + pIndex[lCurrent+1].size - (long)byte_off) / sampsize;

			if (lSamples < 0) lSamples = 1;
		}

		// trim down sample count

		if (lpBuffer && lSamples > cbBuffer / sampsize)
			lSamples = cbBuffer / sampsize;

		if (lStart+lSamples > length)
			lSamples = length - lStart;

		bytecnt = lSamples * sampsize;

		// begin reading frames from this point on

		if (lpBuffer) {
			long actual_bytes = parent->ReadData(
				Stream_Number, lpBuffer,
				pIndex[lCurrent].pos, (int)byte_off,
				bytecnt, &pIndex[lCurrent]);

			if (actual_bytes < sampsize) {
				if (plBytes) *plBytes = 0;
				if (plSamples) *plSamples = 0;
				return AVIERR_FILEREAD;
			}

			actual_bytes -= actual_bytes % sampsize;

			if (plBytes) *plBytes = actual_bytes;
			if (plSamples) *plSamples = actual_bytes / sampsize;

		} else {
			if (plBytes) *plBytes = bytecnt;
			if (plSamples) *plSamples = lSamples;
		}

	} else {
		// Video
		long framesize = pIndex[lStart].size;

		if (lpBuffer) {

			if (framesize > cbBuffer)
				return AVIERR_BUFFERTOOSMALL;

			// read data

			lActual = parent->ReadData(Stream_Number, lpBuffer,
				pIndex[lStart].pos, 0,
				framesize, &pIndex[lStart]);

			if (lActual != framesize) {
				if (plBytes) *plBytes = 0;
				if (plSamples) *plSamples = 0;
				return AVIERR_FILEREAD;
			}
		}

		if (plBytes) *plBytes = framesize;
		if (plSamples) *plSamples = 1;
	}

	return AVIERR_OK;
}

VDPosition ASFReadStream::Start() {
	return 0;
}

VDPosition ASFReadStream::End() {
	return length;
}

long ASFReadStream::PrevKeyFrame(long lFrame) {
	if (lFrame > 0 && lFrame < length) {
		if (psnData->hdr.fccType == streamtypeAUDIO)
			return (lFrame - 1);

		while (--lFrame >= 0)
			if (pIndex[lFrame].flags & ASI_FLAG_KEYFRAME) return lFrame;
	}
	return -1;
}

long ASFReadStream::NextKeyFrame(long lFrame) {
	if (lFrame >= 0 && lFrame < (length - 1)) {
		if (psnData->hdr.fccType == streamtypeAUDIO)
			return (lFrame + 1);

		while (++lFrame < length)
			if (pIndex[lFrame].flags & ASI_FLAG_KEYFRAME) return lFrame;
	}
	return -1;
}

long ASFReadStream::NearestKeyFrame(long lFrame) {
	long ret = 0;

	if (lFrame >= 0 && lFrame < length) {

		if (psnData->hdr.fccType == streamtypeAUDIO	|| IsKeyFrame(lFrame)) {
			ret = lFrame;

		} else {
			ret = PrevKeyFrame(lFrame);
			if (ret < 0) ret = 0;
		}
	}

	return ret;
}

HRESULT ASFReadStream::FormatSize(long lFrame, long *plSize) {
	*plSize = psnData->lFormatLen;
	return AVIERR_OK;
}

HRESULT ASFReadStream::ReadFormat(long lFrame, void *pFormat, long *plSize) {
	if (!pFormat) {
		*plSize = psnData->lFormatLen;
		return 0;
	}

	if (*plSize < psnData->lFormatLen) {
		memcpy(pFormat, psnData->pFormat, *plSize);
	} else {
		memcpy(pFormat, psnData->pFormat, psnData->lFormatLen);
		*plSize = psnData->lFormatLen;
	}

	return AVIERR_OK;
}

bool ASFReadStream::isStreaming() {
	return false;
}

bool ASFReadStream::isKeyframeOnly() {
	return false;
}

bool ASFReadStream::getVBRInfo(double& bitrate_mean, double& bitrate_stddev, double& maxdev) {
	return false;
}

sint64 ASFReadStream::getSampleBytePosition(VDPosition sample_num) {

	if (sample_num < 0 || sample_num >= length) return -1;

	if (psnData->hdr.fccType == streamtypeAUDIO) {
		long lStart = (long)sample_num;
		long lSamples = 1;

		// Handle repeating and/or scrambled audio
		UnscrambleAudio(lStart, lSamples);

		sint64 byte_off = (sint64)lStart * psnData->hdr.dwSampleSize;
		sint64 block_pos;
		long cur;

		// find the frame that has the starting sample -- try and work
		// from our last position to save time

		if (byte_off >= i64Pos) {
			block_pos = i64Pos;
			cur = lCurrent;
			byte_off -= i64Pos;
		} else {
			block_pos = 0;
			cur = 0;
		}

		while (byte_off >= pIndex[cur].size) {
			byte_off -= pIndex[cur].size;
			block_pos += pIndex[cur].size;
			cur++;
		}

		return ((pIndex[cur].pos + byte_off) & 0x0000FFFFFFFFFFFF);	// mask off file field
	}

	// Video
	return (pIndex[sample_num].pos & 0x0000FFFFFFFFFFFF);	// mask off file field
}

#ifdef ASF_VBR_FIX
long ASFReadStream::TimeToSample(long lTimeInMs) {

	// Start with an educated guess
	long guess = VDRoundToLong(
		(lTimeInMs * (double)psnData->hdr.dwRate) / (psnData->hdr.dwScale * 1000.0)
	);

	// Clamp...
	if (guess >= length) guess = length - 1;
	if (guess < 0) guess = 0;

	// Refine (brute force)...
	if (pIndex[guess].pres_time > lTimeInMs) {

		// Look for an earlier sample
		while (guess > 0) {
			--guess;
			if (pIndex[guess].pres_time <= lTimeInMs) {
				// Choose the closer of the two
				const long diff1 = lTimeInMs - pIndex[guess].pres_time;
				const long diff2 = pIndex[guess+1].pres_time - lTimeInMs;
				if (diff2 < diff1) ++guess;
				break;
			}
		}

	} else if (pIndex[guess].pres_time < lTimeInMs) {

		// Look for a later sample
		while (guess < (length - 1)) {
			++guess;
			if (pIndex[guess].pres_time >= lTimeInMs) {
				// Choose the closer of the two
				const long diff1 = pIndex[guess].pres_time - lTimeInMs;
				const long diff2 = lTimeInMs - pIndex[guess-1].pres_time;
				if (diff2 < diff1) --guess;
				break;
			}
		}
	}

	return guess;
}
#endif	// ASF_VBR_FIX

VDPosition ASFReadStream::TimeToPosition(VDTime timeInUs) {
#ifdef ASF_VBR_FIX
	if (psnData->hdr.fccType == streamtypeAUDIO) {
		return (VDPosition)TimeToSample(VDRoundToLong((double)timeInUs / 1000.0));
	}
#endif
	return VDRoundToInt64((timeInUs * (double)psnData->hdr.dwRate) / (psnData->hdr.dwScale * 1000000.0));
}

VDTime ASFReadStream::PositionToTime(VDPosition sample) {
	if (sample >= 0 && sample < length) {
#ifdef ASF_VBR_FIX
		if (psnData->hdr.fccType == streamtypeAUDIO) {
			return (VDTime)(pIndex[sample].pres_time * 1000I64);
		}
#endif
		return VDRoundToInt64((sample * psnData->hdr.dwScale * 1000000.0) / (double)psnData->hdr.dwRate);
	}
	return 0;
}

///////////////////////////////////////////////////////////////////////////


ASFReadHandler::ASFReadHandler(const wchar_t *s, bool _fQuiet, bool _fAcceptPartial)
	: fQuiet		(_fQuiet)
	, fAcceptPartial(_fAcceptPartial) {

	hFile = INVALID_HANDLE_VALUE;
	read_buffer = NULL;
	pktbuf = NULL;
	ref_count = 1;

	mTextInfoCodePage = 0;
	mTextInfoCountryCode = 0;
	mTextInfoLanguage = 0;
	mTextInfoDialect = 0;

	if (!g_disklockinited) {
		g_disklockinited = true;
		InitializeCriticalSection(&g_diskcs);
	}

	_construct(s);
}

ASFReadHandler::~ASFReadHandler() {
	_destruct();
}

void ASFReadHandler::_construct(const wchar_t *pszFile) {

	try {
		// Open file

		if (!(read_buffer = new unsigned char[READ_BUFFER_SIZE]))
			throw MyMemoryError();

		if (GetVersion() & 0x80000000)
			hFile = CreateFileA(VDTextWToA(pszFile).c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		else
			hFile = CreateFileW(pszFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

		if (hFile == INVALID_HANDLE_VALUE) {
			throw MyWin32Error("Couldn't open %s: %%s", GetLastError(), VDTextWToA(pszFile).c_str());
		}

		// Get file size
		{
			DWORD dwHi;
			DWORD dwLo = GetFileSize(hFile, &dwHi);

			if (dwLo == 0xFFFFFFFF) {
				DWORD gle = GetLastError();
				if (gle != NO_ERROR) {
					throw MyWin32Error("Couldn't get file size of %s: %%s", gle, VDTextWToA(pszFile).c_str());
				}
			}

			i64Size = ((__int64)dwHi << 32) + dwLo;
		}

		if (i64Size <= 0) throw MyError("Invalid file size: %I64i", i64Size);

		i64FilePosition = 0;
		_refill_buffer();

		// Parse file

		_parseFileRoot();

	} catch(...) {

		_destruct();
		throw;
	}
}

void ASFReadHandler::_destruct() {
	ASFStreamNode *pasn;

	while(pasn = (ASFStreamNode *)listStreams.RemoveTail())
		delete pasn;

	if (hFile != INVALID_HANDLE_VALUE) {
		CloseHandle(hFile);
		hFile = INVALID_HANDLE_VALUE;
	}

	delete [] read_buffer;
	read_buffer = NULL;

	delete [] pktbuf;
	pktbuf = NULL;
}

void ASFReadHandler::_parseFileRoot() {

	ASF_Object hdr;

	__int64 i64MoviePos = 0;
	__int64 i64MovieLen;

	pktsize = 0;

	// ASF Header Object is mandatory

	if (_readFile(&hdr, 24) >= 24) {
		if (hdr.Object_ID == GUID_ASF_Header_Object && hdr.Object_Size
			&& (_posFile() + hdr.Object_Size - 24) < i64Size) {

			_parse_ASF_Header_Object(hdr.Object_Size - 24);
		}
	}

	if (!pktsize)
		throw MyError("ASF: Header Object is invalid");

	if (_readFile(&hdr, 24) > 0) {
		if (hdr.Object_ID == GUID_ASF_Data_Object) {
			i64MoviePos = _posFile();
			i64MovieLen = std::min<__int64>(hdr.Object_Size - 24, i64Size - i64MoviePos);
			if (i64MovieLen < (hdr.Object_Size - 24)) {
				VDLogAppMessage(kVDLogWarning, kVDST_InputFileASF, kVDM_DataObjectTruncated);
			}
		}
	}

	// Indices may follow; should we parse them?

	if (!i64MoviePos)
		throw MyError("ASF: Data Object not found");

	i64MovieLimit = i64MoviePos + i64MovieLen;

	// Allocate packet buffer

	if (!(pktbuf = new BYTE[pktsize + 16]))
		throw MyMemoryError();

	_parse_ASF_Data_Object(i64MoviePos, i64MovieLen);

	// Glue all indices together and deduce frame speeds

	{
		ASFStreamNode *pasn, *pasn_next;

		pasn = (ASFStreamNode *)listStreams.AtHead();

		while(pasn_next = (ASFStreamNode *)pasn->NextFromHead()) {
			// Make an array of the sample infos
			pasn->sample_list = (ASFSampleInfo *)pasn->dv->MakeArray();
			if (!pasn->sample_list) throw MyMemoryError();

			pasn->lLength = pasn->dv->Length();

			// Count the total bytes in the stream
			pasn->bytes = 0;
			for (int i = 0; i < pasn->lLength; i++) {
				pasn->bytes += pasn->sample_list[i].size;
			}

			// Set stream header length
			if (pasn->hdr.fccType == streamtypeAUDIO) {

				pasn->hdr.dwLength = (DWORD)(pasn->bytes / pasn->hdr.dwSampleSize);
				pasn->lLength = (long)pasn->hdr.dwLength;

			} else {
				const long i = pasn->lLength;
				pasn->hdr.dwLength = i;

				// Guess the framerate
				if (i >= 2) {
					pasn->hdr.dwRate = (DWORD)(
						(double)((i - 1) * 1000000.0) /
						(double)(pasn->sample_list[i-1].pres_time
						- pasn->sample_list[0].pres_time)
					);
				} else {
					// Only one frame?
					pasn->hdr.dwRate = 15000;
				}
				pasn->hdr.dwScale = 1000;

				mFrameRate = VDFraction(pasn->hdr.dwRate, pasn->hdr.dwScale);
			}

			// We can delete the stream's DataVector now
			delete pasn->dv;
			pasn->dv = NULL;

			pasn = pasn_next;
		}
	}
}

void ASFReadHandler::_parse_ASF_Header_Object(__int64 len) {

	// ASF Header Object is the only object which can
	// contain other objects.  It must contain the following
	// to be valid:
	//		ASF File Properties Object
	//		ASF Stream Properties Object

	ASF_Header_Object asfhdr;
	__int64 pos_start;
	
	if (_readFile(&asfhdr, sizeof asfhdr) >= sizeof asfhdr) {
		ASF_Object obj;

		while (asfhdr.Number_Of_Header_Objects-- > 0) {

VDDEBUG("Number_Of_Header_Objects = %ld\n", asfhdr.Number_Of_Header_Objects);

			pos_start = _posFile();
			_readFile(&obj, 24);

VDDEBUG("obj.Object_Size = %I64i\n", obj.Object_Size);

			if (obj.Object_Size < 24 ||
				(_posFile() + obj.Object_Size - 24) > i64Size)
					break;

VDDEBUG("obj.Object_Size OK\n");

			if (obj.Object_ID == GUID_ASF_File_Properties_Object) {
				_parse_ASF_File_Properties_Object(obj.Object_Size - 24);

			} else if (obj.Object_ID == GUID_ASF_Stream_Properties_Object) {
				_parse_ASF_Stream_Properties_Object(obj.Object_Size - 24);

			} else if (obj.Object_ID == GUID_ASF_Content_Description_Object) {
				_parse_ASF_Content_Description_Object(obj.Object_Size - 24);

			} else if (obj.Object_ID == GUID_ASF_Header_Extension_Object) {
				_parse_ASF_Header_Extension_Object(obj.Object_Size - 24);

			}

			_seekFile(pos_start + obj.Object_Size);
		}
	}
}

void ASFReadHandler::_parse_ASF_Header_Extension_Object(__int64 len) {
	__int64 pos_start, pos_max;
	ASF_Header_Extension_Object hdr;

	pos_max = _posFile() + len;
	_readFile(&hdr, sizeof(hdr));

	while ((pos_start = _posFile()) < pos_max) {
		ASF_Object obj;

		_readFile(&obj, 24);

		if (obj.Object_ID == GUID_ASF_Extended_Stream_Properties_Object) {
			_parse_ASF_Extended_Stream_Properties_Object(obj.Object_Size - 24);
		}

		_seekFile(pos_start + obj.Object_Size);
	}
}

void ASFReadHandler::_parse_ASF_Extended_Stream_Properties_Object(__int64 len) {
	__int64 start, rem;
	ASF_Extended_Stream_Properties_Object hdr;
	int i;

	start = _posFile();

	_readFile(&hdr, sizeof(hdr));

	for (i = 0; i < hdr.Stream_Name_Count; i++) {
		ASF_Stream_Name sn_hdr;

		_readFile(&sn_hdr, sizeof(sn_hdr));

		if (sn_hdr.Stream_Name_Length != 0) {
			_seekFile(_posFile() + sn_hdr.Stream_Name_Length);
		}
	}

	for (i = 0; i < hdr.Payload_Extension_System_Count; i++) {
		ASF_Payload_Extension_System pes_hdr;

		_readFile(&pes_hdr, sizeof(pes_hdr));

		if (pes_hdr.Extension_System_Info_Length != 0) {
			_seekFile(_posFile() + pes_hdr.Extension_System_Info_Length);
		}
	}

	rem = len - (_posFile() - start);
	if (rem >= sizeof(ASF_Stream_Properties_Object)) {
		ASF_Object obj;

		_readFile(&obj, 24);

		_parse_ASF_Stream_Properties_Object(obj.Object_Size - 24);
	}

	rem = len - (_posFile() - start);
	if (rem != 0) {
		throw MyError("ASF Extended Stream Properties Object has invalid size");
	}
}

void ASFReadHandler::_parse_ASF_File_Properties_Object(__int64 len) {
	ASF_File_Properties_Object hdr;

	if (len != sizeof hdr)
		throw MyError("ASF File Properties Object has invalid size (%ld bytes)", len);

	_readFile(&hdr, (long)len);

	lPreroll = (long)hdr.Preroll;

	{
		SYSTEMTIME st;
		char sa[64];
		int i = 0;

		FileTimeToSystemTime((FILETIME *)(&(hdr.Creation_Date)), &st);

		switch (st.wMonth) {
			case 1: i = sprintf(sa, "January"); break;
			case 2: i = sprintf(sa, "February"); break;
			case 3: i = sprintf(sa, "March"); break;
			case 4: i = sprintf(sa, "April"); break;
			case 5: i = sprintf(sa, "May"); break;
			case 6: i = sprintf(sa, "June"); break;
			case 7: i = sprintf(sa, "July"); break;
			case 8: i = sprintf(sa, "August"); break;
			case 9: i = sprintf(sa, "September"); break;
			case 10: i = sprintf(sa, "October"); break;
			case 11: i = sprintf(sa, "November"); break;
			case 12: i = sprintf(sa, "December"); break;
		}

		i += sprintf(sa+i, " %d, %d", st.wDay, st.wYear);
		mTextInfo.push_back(tTextInfo::value_type('DRCI', VDStringA(sa)));
	}


	// Note: The original used Minimum_Data_Packet_Size
	pktsize = hdr.Maximum_Data_Packet_Size;
}

/*
	case 'LRAI':	// Archival Location
	case 'TRAI':	// Artist (Author)
	case 'SMCI':	// Commissioned By
	case 'TMCI':	// Comments
	case 'POCI':	// Copyright
	case 'DRCI':	// Creation Date
	case 'PRCI':	// Cropped
	case 'MIDI':	// Dimensions
	case 'IPDI':	// Dots Per Inch
	case 'GNEI':	// Engineer
	case 'RNGI':	// Genre
	case 'YEKI':	// Keywords
	case 'TGLI':	// Lightness
	case 'DEMI':	// Medium
	case 'MANI':	// Name
	case 'TLPI':	// Palette Setting
	case 'DRPI':	// Product
	case 'JBSI':	// Subject
	case 'TFSI':	// Software
	case 'PHSI':	// Sharpness
	case 'CRSI':	// Source
	case 'FRSI':	// Source Form
	case 'HCTI':	// Technician
*/

void ASFReadHandler::_parse_ASF_Content_Description_Object(__int64 len) {
	const int maximum = 1024;

	int Title_Length = 0;
	int Author_Length = 0;
	int Copyright_Length = 0;
	int Description_Length = 0;
	int Rating_Length = 0;

	wchar_t *sw;
	int slen;

	__int64 pos_start;

	if (len < 10)
		throw MyError("ASF: Content Description Object is too short");

	_readFile(&Title_Length, 2);
	_readFile(&Author_Length, 2);
	_readFile(&Copyright_Length, 2);
	_readFile(&Description_Length, 2);
	_readFile(&Rating_Length, 2);

	// Aaargh... Must we convert to ANSI?

	if (!(sw = new wchar_t[maximum]))
		throw MyMemoryError();

	if (Title_Length > 1) {

		slen = Title_Length >> 1;
		if (slen >= maximum) slen = maximum - 1;

		pos_start = _posFile();
		_readFile(sw, slen * 2);
		sw[slen] = 0;

		mTextInfo.push_back(tTextInfo::value_type('MANI', VDStringA(VDTextWToA(sw).c_str())));

		_seekFile(pos_start + Title_Length);
	}

	if (Author_Length > 1) {

		slen = Author_Length >> 1;
		if (slen >= maximum) slen = maximum - 1;

		pos_start = _posFile();
		_readFile(sw, slen * 2);
		sw[slen] = 0;

		mTextInfo.push_back(tTextInfo::value_type('TRAI', VDStringA(VDTextWToA(sw).c_str())));

		_seekFile(pos_start + Author_Length);
	}

	if (Copyright_Length > 1) {

		slen = Copyright_Length >> 1;
		if (slen >= maximum) slen = maximum - 1;

		pos_start = _posFile();
		_readFile(sw, slen * 2);
		sw[slen] = 0;

		mTextInfo.push_back(tTextInfo::value_type('POCI', VDStringA(VDTextWToA(sw).c_str())));

		_seekFile(pos_start + Copyright_Length);
	}

	if (Description_Length > 1) {

		slen = Description_Length >> 1;
		if (slen >= maximum) slen = maximum - 1;

		pos_start = _posFile();
		_readFile(sw, slen * 2);
		sw[slen] = 0;

		mTextInfo.push_back(tTextInfo::value_type('TMCI', VDStringA(VDTextWToA(sw).c_str())));

		_seekFile(pos_start + Description_Length);
	}

	// I haven't decided what to do with "Rating"...

	delete [] sw;
}

void ASFReadHandler::_parse_ASF_Stream_Properties_Object(__int64 len) {

VDDEBUG("Reading ASF_Stream_Properties_Object\n");

	ASF_Stream_Properties_Object hdr;
	ASFStreamNode *pasn;

	unsigned char *Type_Specific_Data = NULL;
	unsigned char *Error_Correction_Data = NULL;

	if (len < sizeof hdr)
		throw MyError("ASF: stream header is too short");

	_readFile(&hdr, sizeof hdr);

	if (hdr.Type_Specific_Data_Length > 0) {
		if (!(Type_Specific_Data = new unsigned char[hdr.Type_Specific_Data_Length]))
			throw MyMemoryError();

		_readFile(Type_Specific_Data, hdr.Type_Specific_Data_Length);
	}

	if (hdr.Error_Correction_Data_Length > 0) {
		if (!(Error_Correction_Data = new unsigned char[hdr.Error_Correction_Data_Length]))
			throw MyMemoryError();

		_readFile(Error_Correction_Data, hdr.Error_Correction_Data_Length);
	}

	// is it a type we know?

	if (	hdr.Stream_Type == GUID_ASF_Video_Media
		||	hdr.Stream_Type == GUID_ASF_Audio_Media) {

		if (hdr.Flags & 0x8000) {
			if (hdr.Stream_Type == GUID_ASF_Video_Media)
				throw MyError("ASF: Video Stream is encrypted");
			else
				throw MyError("ASF: Audio Stream is encrypted");
		}

		// Create a new stream node
		if (!(pasn = new ASFStreamNode()))
			throw MyMemoryError();

		if (!(pasn->dv = new DataVector(sizeof ASFSampleInfo)))
			throw MyMemoryError();

		pasn->stream_number = hdr.Flags & 0x007F;

		memset(&pasn->hdr, 0, sizeof(AVISTREAMINFO));
		pasn->hdr.dwQuality = -1;

		if (hdr.Stream_Type == GUID_ASF_Video_Media) {
			// Video Stream, starts with:
			//	DWORD	Encoded Image Width
			//	DWORD	Encoded Image Height
			//	BYTE	Reserved Flags
			//	WORD	Format Data Size

			BITMAPINFOHEADER *bmih = (BITMAPINFOHEADER *)(Type_Specific_Data+11);

			_RPT0(0,"ASF: video stream detected\n");

			pasn->hdr.fccType		= streamtypeVIDEO;
			pasn->hdr.fccHandler	= bmih->biCompression;
			pasn->hdr.dwRate		= 1000000;
			pasn->hdr.dwScale		= 33367;
			pasn->hdr.dwSampleSize	= 0;

			pasn->lFormatLen = *((WORD *)(Type_Specific_Data+9));

			if (!(pasn->pFormat = new char[pasn->lFormatLen]))
				throw MyMemoryError();

			bmih->biSize = pasn->lFormatLen;
			memcpy(pasn->pFormat, bmih, pasn->lFormatLen);

		} else {
			// Audio Stream
			WAVEFORMATEX *wfex = (WAVEFORMATEX *)Type_Specific_Data;

			_RPT0(0,"ASF: audio stream detected\n");

			pasn->hdr.fccType		= streamtypeAUDIO;
			pasn->hdr.fccHandler	= NULL;
			pasn->hdr.dwRate		= wfex->nAvgBytesPerSec;
			pasn->hdr.dwScale		= wfex->nBlockAlign;
			pasn->hdr.dwSampleSize	= wfex->nBlockAlign;

			// Patch?

// These GUIDs are mysterious:
//	1A0F78F0-EC8A-11d2-BBBE-006008320064
//	F6DC9830-BC79-11d2-A9D0-006097926036

#ifdef PATCH_WMA
		// If it's Microsoft Audio Codec, manually patch the format

			if (wfex->wFormatTag == 0x161) {	// Windows Media Audio v2
				pasn->lFormatLen = sizeof(WAVEFORMATEX)+10+37;

				if (!(pasn->pFormat = new char[pasn->lFormatLen]))
					throw MyMemoryError();

				memcpy(pasn->pFormat, wfex, sizeof(WAVEFORMATEX)+10);

				strcpy((char *)pasn->pFormat+18+10, "1A0F78F0-EC8A-11d2-BBBE-006008320064");

				((WAVEFORMATEX *)pasn->pFormat)->cbSize = 0x2f;

			} else if (wfex->wFormatTag == 0x160) { // Windows Media Audio v1

				pasn->lFormatLen = sizeof(WAVEFORMATEX)+4+37;

				if (!(pasn->pFormat = new char[pasn->lFormatLen]))
					throw MyMemoryError();

				memcpy(pasn->pFormat, wfex, sizeof(WAVEFORMATEX)+4);

				strcpy((char *)pasn->pFormat+18+4, "1A0F78F0-EC8A-11d2-BBBE-006008320064");

				((WAVEFORMATEX *)pasn->pFormat)->cbSize = 0x29;

			} else {
#endif
				pasn->lFormatLen = hdr.Type_Specific_Data_Length;

				if (!(pasn->pFormat = new char[pasn->lFormatLen]))
					throw MyMemoryError();

				memcpy(pasn->pFormat, wfex, pasn->lFormatLen);

#ifdef PATCH_WMA
			}
#endif
			if (hdr.Error_Correction_Type == GUID_ASF_Audio_Spread
				&& hdr.Error_Correction_Data_Length >= 8) {

				const unsigned short *ecc =
					(unsigned short *)(Error_Correction_Data+1);

				// Span: the number of packets over
				// which audio will be spread (usually 1)
				pasn->iSpan = Error_Correction_Data[0];

				// Virtual Pack Length / Virtual Chunk Length
				// (This result should always be 1)
				if (ecc[1]) {
					pasn->lAudioScrambleChunk = ecc[0] / ecc[1];
				} else {
					pasn->lAudioScrambleChunk = 1;
				}
			}
		}

		listStreams.AddTail(pasn);
	}

	if (Error_Correction_Data) delete [] Error_Correction_Data;
	if (Type_Specific_Data) delete [] Type_Specific_Data;
}


// Depending on the value of "flags", we need to read either
// a one, two, or four byte operand from the data. The return
// value tells the number of bytes read.
// (Ugly hack, I know...)

int ASFReadHandler::_get_VLC(const void *data, DWORD *VLC, int flags) {
	if (data) {
		switch (flags & 3) {

			case 0x01:
				if (VLC) *VLC = *(BYTE *)data;
				return 1;

			case 0x02:
				if (VLC) *VLC = *(WORD *)data;
				return 2;

			case 0x03:
				if (VLC) *VLC = *(DWORD *)data;
				return 4;
		}
	}

	if (VLC) *VLC = 0;

	return 0;
}

void ASFReadHandler::_parse_ASF_Data_Object(__int64 pos, __int64 len) {
	ASFSampleInfo asi_list[128];
	ASF_Data_Object hdr;

	for (int i = 0; i < 128; i++) {
		asi_list[i].flags = 0;
	}

// Pop up parse dialog
	ProgressDialog pd(NULL, "ASF Import Filter", "Constructing ASF index block", (long)((len+1023)/1024), true);
	pd.setValueFormat("%ldK of %ldK");

	_seekFile(pos);
	_readFile(&hdr, sizeof(hdr));
	len -= sizeof(hdr);

// Now we are right at the beginning of the packets

	while (len >= pktsize) {
		BYTE *pp = pktbuf;

		BYTE	Length_Type_Flags;
		BYTE	Property_Flags;
		BYTE	Payload_Flags;
		BYTE	Number_Of_Payloads;
		DWORD	Send_Time;
		DWORD	Packet_Length;
		DWORD	Padding_Length;

		int pktleft;

	// The following reads the whole packet
		_readFile(pp, pktsize);
		len -= pktsize;

		Length_Type_Flags = *pp++;	// Error Correction Flags?
							// 1....... EC Present (1)
							// .xx..... EC Length Type
							// ...x.... Opaque Data Present
							// ....xxxx EC Data Length

		VDASSERT(Length_Type_Flags == 0x82);

		if (Length_Type_Flags & 0x80) {
			int EC_Length_Type = (Length_Type_Flags >> 5) & 3;
			int EC_Data_Length = Length_Type_Flags & 0x0f;

			if (EC_Length_Type == 0) pp += EC_Data_Length;
			Length_Type_Flags = *pp++;
		}

		Property_Flags = *pp++;

		pp += _get_VLC(pp, NULL, Length_Type_Flags >> 1);
		pp += _get_VLC(pp, &Packet_Length, Length_Type_Flags >> 5);
		pp += _get_VLC(pp, &Padding_Length, Length_Type_Flags >> 3);

		Send_Time = *(DWORD *)pp;
		pp += 6;	// skips Duration

		if (Length_Type_Flags & 1) {
			// Multiple payloads
			Payload_Flags = *pp++;

			// Why the heck does this line flag warning C4244 ?!?
			//Number_Of_Payloads = Payload_Flags & 0x3F;

			// Just STFU PLEASE
			Number_Of_Payloads = (BYTE)(Payload_Flags & 0x3F);

		} else {
			// Single payload
			Number_Of_Payloads = 1;
		}

		// At this point we're at the
		// beginning of the payload(s)

		pktleft = (pktsize - (pp - pktbuf)) - Padding_Length;

		// parse down segments/subpackets/whatever...

		// Whether Multiple or Single, right now
		// we are at the "Stream_Number" field

		while (Number_Of_Payloads--) {
			DWORD Media_Object_Number;
			DWORD Offset_Into_Media_Object;
			bool Key_Frame_Bit;
			DWORD Stream_Number;
			DWORD Payload_Length;
			DWORD Replicated_Data_Length;
			DWORD Presentation_Time;
			int i;

			i = _get_VLC(pp, &Stream_Number, Property_Flags >> 6);
			pp += i; pktleft -= i;

			i = _get_VLC(pp, &Media_Object_Number, Property_Flags >> 4);
			pp += i; pktleft -= i;

		// The following is Offset Into Media Object for uncompressed
		// payloads, but it's Presentation Time for compressed payloads:

			i = _get_VLC(pp, &Offset_Into_Media_Object, Property_Flags >> 2);
			pp += i; pktleft -= i;

			i = _get_VLC(pp, &Replicated_Data_Length, Property_Flags);
			pp += i; pktleft -= i;

			Key_Frame_Bit = !!(Stream_Number & 0x80);
			Stream_Number &= 0x7F;

			if (Replicated_Data_Length == 1) {

		// Compressed payload(s), either Multiple or Single
		// (This used to be referred to as "grouping")

		// All payloads contain sub-payloads from a single stream,
		// and each sub-payload is a complete Media Object

				BYTE Presentation_Time_Delta;
				BYTE *pplimit;

				Presentation_Time = Offset_Into_Media_Object;
				Offset_Into_Media_Object = 0;

				Presentation_Time_Delta = *pp++;
				pktleft--;

				if (Length_Type_Flags & 1) {
					// Multiple payloads

					// Payload Length (supposed to be a WORD)
					i = _get_VLC(pp, &Payload_Length, Payload_Flags >> 6);
					pp += i; pktleft -= i;

				} else {
					// Single payload

					Payload_Length = pktleft;
				}

				pktleft -= Payload_Length;
				pplimit = pp + Payload_Length;

				while (pp < pplimit) {
					BYTE Sub_Payload_Data_Length = *pp++;

					pp += Sub_Payload_Data_Length;

					// Add sample to the appropriate stream
					ASFStreamNode *pasn = (ASFStreamNode *)listStreams.AtHead();
					ASFStreamNode *pasn_next;

					while (pasn_next = (ASFStreamNode *)pasn->NextFromHead()) {
						if (pasn->stream_number == Stream_Number) {
							break;
						}
						pasn = pasn_next;
					}

					if (pasn_next != NULL) {
						ASFSampleInfo *asi = &asi_list[Stream_Number];

						asi->pos = _posFile() - pktsize;
						asi->size = Sub_Payload_Data_Length;
						asi->pres_time = Presentation_Time - lPreroll;
						if (asi->pres_time < 0) asi->pres_time = 0;
						asi->obj_number = (BYTE)Media_Object_Number;

						// This generates a stupid warning (aargh...!)
						//asi->flags = Key_Frame_Bit? ASI_FLAG_KEYFRAME: 0;

						if (Key_Frame_Bit) {
							asi->flags = ASI_FLAG_KEYFRAME;
						} else {
							asi->flags = 0;
						}

						pasn->dv->Add(asi);
					}

					Presentation_Time += Presentation_Time_Delta;
					Media_Object_Number = (Media_Object_Number + 1) & 0xFF;
				}


			} else if (Replicated_Data_Length == 0 || Replicated_Data_Length >= 8) {

				DWORD Size_Of_Media_Object;

				// Hmm, I haven't seen this fail yet:
				VDASSERT(Replicated_Data_Length > 0);

			// Uncompressed payload(s), either Multiple or Single
			// (Payloads need not contain an entire Media Object)

				if (Replicated_Data_Length >= 8) {
					asi_list[Stream_Number].size = *(DWORD *)(pp);
					asi_list[Stream_Number].pres_time = *(DWORD *)(pp+4);

					pp += Replicated_Data_Length;
					pktleft -= Replicated_Data_Length;
				}
				Size_Of_Media_Object = asi_list[Stream_Number].size;
				Presentation_Time = asi_list[Stream_Number].pres_time;

				if (Length_Type_Flags & 0x01) {
					// Multiple payloads, uncompressed

					// Payload Length (supposed to be a WORD)
					i = _get_VLC(pp, &Payload_Length, Payload_Flags >> 6);
					pp += i; pktleft -= i;

				} else {

					// Single payload, uncompressed
					Payload_Length = pktleft;
				}

				pp += Payload_Length;
				pktleft -= Payload_Length;

				// Add the object to the appropriate stream.  For fragmented objects, only add it
				// once, and only if it's complete.  Keep in mind that a stream might start with an
				// incomplete object (if recording of a live stream began during a fragmented
				// object after the first fragment had been transmitted).

				if (Offset_Into_Media_Object == 0) {
					ASFSampleInfo *asi = &asi_list[Stream_Number];

					asi->pos = _posFile() - pktsize;
					asi->size = Size_Of_Media_Object;
					asi->pres_time = Presentation_Time - lPreroll;
					if (asi->pres_time < 0) asi->pres_time = 0;
					asi->obj_number = (BYTE)Media_Object_Number;

					//asi->flags = Key_Frame_Bit? ASI_FLAG_KEYFRAME: 0;
					if (Key_Frame_Bit) {
						asi->flags = ASI_FLAG_KEYFRAME | ASI_FLAG_SEEN_FIRST_FRAGMENT;
					} else {
						asi->flags = ASI_FLAG_SEEN_FIRST_FRAGMENT;
					}
				}

				if (((Offset_Into_Media_Object + Payload_Length) >= Size_Of_Media_Object) &&
					(asi_list[Stream_Number].flags & ASI_FLAG_SEEN_FIRST_FRAGMENT))
				{
					ASFStreamNode *pasn = (ASFStreamNode *)listStreams.AtHead();
					ASFStreamNode *pasn_next;

					asi_list[Stream_Number].flags -= ASI_FLAG_SEEN_FIRST_FRAGMENT;

					while (pasn_next = (ASFStreamNode *)pasn->NextFromHead()) {
						if (pasn->stream_number == Stream_Number) {
							break;
						}
						pasn = pasn_next;
					}

					if (pasn_next != NULL) {
						pasn->dv->Add(&asi_list[Stream_Number]);
					}
				}

			} else {

				if (fAcceptPartial) return;

				throw MyError(
					"ASF: Invalid Replicated_Data_Length %d in packet at %I64x\n"
					"packet size = %lx, pp = %lx"
					, (unsigned char)Replicated_Data_Length, _posFile()-pktsize, pktsize, pp-pktbuf);
			}

		}

		// Update parse dialog
		pd.advance((long)((_posFile() - pos)/1024));
		pd.check();
	}

	VDDEBUG("ASF::_parse_ASF_Data_Object finished, len = %I64i\n", len);
}

///////////////////////////////////////////////////////////////////////////

void ASFReadHandler::Release() {
	if (!--ref_count)
		delete this;
}

void ASFReadHandler::AddRef() {
	++ref_count;
}

IAVIReadStream *ASFReadHandler::GetStream(DWORD fccType, LONG lParam) {
	ASFStreamNode *pasn, *pasn_next;
//	int streamno = 0;

	pasn = (ASFStreamNode *)listStreams.AtHead();

	while(pasn_next = (ASFStreamNode *)pasn->NextFromHead()) {
		if ((pasn->hdr.fccType == fccType) && (pasn->lLength > 0))
			break;

		pasn = pasn_next;
	}

	if (pasn_next) {
		return new ASFReadStream(this, pasn);
	}

	return NULL;
}

void ASFReadHandler::EnableFastIO(bool f) {
	fDisableFastIO = !f;
}

bool ASFReadHandler::isOptimizedForRealtime() {
	return false;
}

bool ASFReadHandler::isStreaming() {
	return false;
}

bool ASFReadHandler::isIndexFabricated() {
	return false;
}

bool ASFReadHandler::AppendFile(const wchar_t *pszFile) {
	return false;
}

bool ASFReadHandler::getSegmentHint(const char **ppszPath) {
	if (ppszPath)
		*ppszPath = NULL;
	return false;
}

void ASFReadHandler::GetTextInfo(tTextInfo& textInfo) {
	textInfo = mTextInfo;
}

void ASFReadHandler::GetTextInfoEncoding(int& codePage, int& countryCode, int& language, int& dialect) {
	codePage		= mTextInfoCodePage;
	countryCode		= mTextInfoCountryCode;
	language		= mTextInfoLanguage;
	dialect			= mTextInfoDialect;
}

///////////////////////////////////////////////////////////////////////////


long ASFReadHandler::ReadData(int streamno, void *lpBuffer, __int64 pktpos, int offset, long bytes, ASFSampleInfo *asi) {
	unsigned char current_obj;
	long fragment_point = 0;
	char *lpData = (char *)lpBuffer;

	// FIXME: Any way to bounds check this?
	current_obj = asi->obj_number;
//	asi++;

	if (!bytes)
		return 0;

	_seekFile(pktpos);

	_RPT3(0,"start>> %d, %I64x, %ld\n", bytes, _posFile(), offset);

	while(bytes > 0 && _posFile() < i64MovieLimit) {
		int pktleft;
		BYTE *pp = pktbuf;
		BYTE Length_Type_Flags;
		BYTE Property_Flags;
		BYTE Payload_Flags;
		BYTE Number_Of_Payloads;
		DWORD Padding_Length = 0;

		_readFile(pktbuf, pktsize);

		Length_Type_Flags = *pp++;	// Error Correction Flags?
							// 1....... EC Present (1)
							// .xx..... EC Length Type
							// ...x.... Opaque Data Present
							// ....xxxx EC Data Length

		if (Length_Type_Flags & 0x80) {
			int EC_Length_Type = (Length_Type_Flags >> 5) & 3;
			int EC_Data_Length = Length_Type_Flags & 0x0F;

			if (EC_Length_Type == 0) pp += EC_Data_Length;
			Length_Type_Flags = *pp++;
		}

		Property_Flags = *pp++;

		pp += _get_VLC(pp, NULL, Length_Type_Flags >> 1);
		pp += _get_VLC(pp, NULL, Length_Type_Flags >> 5);
		pp += _get_VLC(pp, &Padding_Length, Length_Type_Flags >> 3);

		pp += 6;	// Skip Send Time and Duration

		if (Length_Type_Flags & 1) {
			// Multiple payloads
			Payload_Flags = *pp++;
			Number_Of_Payloads = (BYTE)(Payload_Flags & 0x3F);
		} else {
			// Single payload
			Number_Of_Payloads = 1;
		}

		// At this point we're at the beginning of the payload(s)

		pktleft = (pktsize - (pp - pktbuf)) - Padding_Length;

		// parse down subpackets

		while (Number_Of_Payloads--) {
			DWORD Offset_Into_Media_Object;
			DWORD Stream_Number;
			DWORD Media_Object_Number;
			DWORD Payload_Length;
			DWORD Replicated_Data_Length;
			BYTE *ppData;
			int i;

			i = _get_VLC(pp, &Stream_Number, Property_Flags >> 6);
			pp += i; pktleft -= i;

			i = _get_VLC(pp, &Media_Object_Number, Property_Flags >> 4);
			pp += i; pktleft -= i;

			i = _get_VLC(pp, &Offset_Into_Media_Object, Property_Flags >> 2);
			pp += i; pktleft -= i;

			i = _get_VLC(pp, &Replicated_Data_Length, Property_Flags);
			pp += i; pktleft -= i;

			Stream_Number &= 0x7F;

			if (Replicated_Data_Length == 1) {

		// Compressed payload(s), either Multiple or Single
		// (This used to be referred to as "grouping")

		// All payloads contain sub-payloads from a single stream,
		// and each sub-payload is a complete Media Object

				Offset_Into_Media_Object = 0;

				pp++;	// Skip Presentation Time Delta
				pktleft--;

				if (Length_Type_Flags & 1) {

					i = _get_VLC(pp, &Payload_Length, Payload_Flags >> 6);
					pp += i; pktleft -= i;

				} else {
					Payload_Length = pktleft;
				}

				ppData = pp;

				pp += Payload_Length;
				pktleft -= Payload_Length;

				if (Stream_Number == streamno) {
					BYTE Sub_Payload_Data_Length;

				// Since this is a grouped set, if we are still in the middle of fragment
				// assembly we've missed a fragment somewhere.

					if (fragment_point)
						throw MyError("ASF: missing fragment (group, fragpt=%lx, current_obj=%x, pkt=%I64x)", fragment_point, current_obj, _posFile()-pktsize);

					while (ppData < pp && bytes > 0) {
						Sub_Payload_Data_Length = *ppData++;

						if ((signed char)(Media_Object_Number - current_obj) > 0)
							throw MyError("ASF: missing object (Media_Object_Number=%02x, current_obj=%02x)", Media_Object_Number, current_obj);

						if ((BYTE)Media_Object_Number == current_obj) {

							// Finished with this object, goto next object
							asi++;
							current_obj = asi->obj_number;

							if (offset >= Sub_Payload_Data_Length) {
								// skip this object
								offset -= Sub_Payload_Data_Length;
							} else {
								// fetch the object
								int tc = Sub_Payload_Data_Length - offset;
								if (tc > bytes) tc = bytes;

								memcpy(lpData, ppData + offset, tc);
								lpData += tc;
								bytes -= tc;
								offset = 0;
							}
						}

						Media_Object_Number = (Media_Object_Number + 1) & 0xFF;
						ppData += Sub_Payload_Data_Length;
					}
				}

			} else if (Replicated_Data_Length >= 8) {
				DWORD Size_Of_Media_Object;

				// Uncompressed payload(s), Single or Multiple

				Size_Of_Media_Object = *(DWORD *)pp;

				pp += Replicated_Data_Length;
				pktleft -= Replicated_Data_Length;

				if (Length_Type_Flags & 1) {
					// Multiple payloads, uncompressed

					// Payload Length (supposed to be a WORD)
					i = _get_VLC(pp, &Payload_Length, Payload_Flags >> 6);
					pp += i; pktleft -= i;

				} else {
					// Single payload, uncompressed
					Payload_Length = pktleft;
				}

				ppData = pp;

				pp += Payload_Length;
				pktleft -= Payload_Length;

				// Is it what we want?

				if (Stream_Number == streamno) {

					// A packet might hold objects earlier than the ones we want, so we
					// can ignore those.  But if we hit one later than the one we
					// want, we've gone too far.  Gotta watch out for the FF->00 rollover.
					//
					// Watch out, #'s may not be sequential; EVA2.ASF is missing the 0x65
					// sequence at 0x39796-0x39f16...

					if ((signed char)(Media_Object_Number - current_obj) > 0)
						throw MyError("ASF: missing fragment (Media_Object_Number=%02x, current_obj=%02x, pktpos=%I64x)", Media_Object_Number, current_obj, _posFile()-pktsize);

					if ((BYTE)Media_Object_Number == current_obj) {

						// If we are reassembling fragments, make sure we haven't missed one.
						if (Offset_Into_Media_Object > (DWORD)fragment_point)
							throw MyError("ASF: fragment reassembly error (fragpt=%08lx, should be %08lx)", Offset_Into_Media_Object, fragment_point);

						// Update fragment point
						fragment_point = Offset_Into_Media_Object + Payload_Length;

						if (offset < fragment_point) {
							// We need this fragment
							int tc;
							int frag_off = Offset_Into_Media_Object - offset;

							if (frag_off >= 0) {
								// We want the whole fragment
								tc = bytes - frag_off;
								if (tc > 0) {
									if (tc > Payload_Length) tc = Payload_Length;
									memcpy(lpData + frag_off, ppData, tc);
								}
							} else {
								// We just want a latter portion
								tc = Payload_Length + frag_off;
								if (tc > 0) {
									if (tc > bytes) tc = bytes;
									memcpy(lpData, ppData - frag_off, tc);
								}
							}
						}

						if (fragment_point >= Size_Of_Media_Object) {
							// Finished with this object
							lpData += Size_Of_Media_Object;
							offset -= Size_Of_Media_Object;
							bytes -= Size_Of_Media_Object;
							fragment_point = 0;

							if (offset < 0) offset = 0;

							asi++;
							current_obj = asi->obj_number;
						}
					}
				}
			} else {
				throw MyError("ASF: Unhandled Replicated_Data_Length %02x in packet at %I64lx", (unsigned char)Replicated_Data_Length, _posFile()-pktsize);
			}

			if (bytes <= 0) break;
		}
	}

	_RPT3(0,"end>> %d, %I64x, %ld\n", bytes, _posFile(), offset);

	return lpData - (char *)lpBuffer;
}


///////////////////////////////////////////////////////////////////////////

long ASFReadHandler::_refill_buffer() {
    if (hFile != INVALID_HANDLE_VALUE) {
        DWORD dwActual;

        if (ReadFile(hFile, read_buffer, READ_BUFFER_SIZE, &dwActual, NULL) == 0) dwActual = 0;
        if (dwActual < READ_BUFFER_SIZE) memset(read_buffer+dwActual, 0, READ_BUFFER_SIZE - dwActual);
		return (long)dwActual;
    }
	return 0;
}

long ASFReadHandler::_readFile(void *data, long len) {
	long lActual = 0;

	if (hFile != INVALID_HANDLE_VALUE && len > 0) {
		unsigned char *dst = (unsigned char *)data;
		long offset = (long)(i64FilePosition & (READ_BUFFER_SIZE-1));
		//lActual = len;
		long remains = READ_BUFFER_SIZE - offset;

		for (;;) {
			//long remains = READ_BUFFER_SIZE - offset;
			if (remains > len) remains = len;
			memcpy(dst, read_buffer + offset, remains);
			i64FilePosition += remains;
			lActual += remains;
			len -= remains;
			if (len <= 0) break;
			dst += remains;
			remains = _refill_buffer();
			if (remains <= 0) break;
			offset = 0;
		}

// These conditionals sometimes give different results!?!
//		if (i64FilePosition & (READ_BUFFER_SIZE-1) == 0)
		if ((i64FilePosition & (READ_BUFFER_SIZE-1)) == 0)

			_refill_buffer();
    }

	return lActual;
}
/*
void ASFReadHandler::_readFile2(void *data, long len) {
	long lActual = _readFile(data, len);

	if (lActual < 0)
		throw MyWin32Error("Failure reading ASF file: %%s.",GetLastError());

	if (lActual != len)
		throw MyError("Failure reading ASF file: Unexpected end of file");
}
*/
void ASFReadHandler::_seekFile(__int64 i64NewPos) {
	if (hFile != INVALID_HANDLE_VALUE) {
		if (i64NewPos < 0)
			i64NewPos = 0;
		else if (i64NewPos > i64Size)
			i64NewPos = i64Size;

		__int64 mask1 = i64NewPos & -READ_BUFFER_SIZE;
		__int64 mask2 = i64FilePosition & -READ_BUFFER_SIZE;

		if (mask1 != mask2) {
			LONG lHi = (LONG)(mask1 >> 32);
			DWORD lLo = (DWORD)(mask1 & 0xFFFFFFFF);

			SetFilePointer(hFile, lLo, &lHi, FILE_BEGIN);
			_refill_buffer();
		}
		i64FilePosition = i64NewPos;
    }
}

__int64 ASFReadHandler::_posFile() {
	return i64FilePosition;
}

///////////////////////////////////////////////////////////////////////

class InputFileASF : public InputFileAVI {
private:
	ASFReadHandler *pASFFile;
	ASFReadStream *pASFStreamAudio;
	ASFReadStream *pASFStreamVideo;

	bool fKillAudio;

	FOURCC fccForceVideo;
	FOURCC fccForceVideoHandler;

	static char szME[];

public:
	InputFileASF();
	~InputFileASF();

	void Init(const wchar_t *szFile);
//	void InitStriped(const char *szFile);
	bool Append(const wchar_t *szFile);

	void GetTextInfo(tFileTextInfo& info);

	bool isOptimizedForRealtime();
	bool isStreaming();
/*
	void setOptions(InputFileOptions *_ifo);
	InputFileOptions *createOptions(const char *buf);
	InputFileOptions *promptForOptions(HWND hwnd);
	void EnableSegmentAutoscan();
	void ForceCompatibility();
	void setAutomated(bool fAuto);

	void InfoDialog(HWND hwndParent);
*/
};

char InputFileASF::szME[]="Windows Media Import Filter";

InputFileASF::InputFileASF() {

	pASFFile = NULL;
	pASFStreamVideo = NULL;
	pASFStreamAudio = NULL;
}

InputFileASF::~InputFileASF() {
	audioSrc = NULL;
	videoSrc = NULL;

	if (pASFFile)
		pASFFile->Release();
}

void InputFileASF::Init(const wchar_t *szFile) {
	VDLogAppMessage(kVDLogMarker, kVDST_InputFileASF, kVDM_OpeningFile, 1, &szFile);

	AddFilename(szFile);

	if (!(pASFFile = (ASFReadHandler *)CreateASFReadHandler(szFile, false, false)))
		throw MyMemoryError();

	if (!(videoSrc = new VideoSourceAVI(pASFFile, NULL, NULL, false, false, 0, 0)))
		throw MyMemoryError();

	if (!videoSrc->init())
		throw MyError("%s: problem opening video stream", szME);

	audioSrc = new AudioSourceAVI(pASFFile, false);
	if (!audioSrc->init()) audioSrc = NULL;
}

bool InputFileASF::Append(const wchar_t *szFile) {
	return false;
}

void InputFileASF::GetTextInfo(tFileTextInfo& info) {
	if (pASFFile) pASFFile->GetTextInfo(info);
}

bool InputFileASF::isOptimizedForRealtime() {
	if (pASFFile) return pASFFile->isOptimizedForRealtime();
	return false;
}

bool InputFileASF::isStreaming() {
	if (pASFFile) return pASFFile->isStreaming();
	return false;
}


///////////////////////////////////////////////////////////////////////

class VDInputDriverASF : public vdrefcounted<IVDInputDriver> {
public:
	const wchar_t *GetSignatureName() {
		return L"Windows Media input driver (internal)";
	}

	uint32 GetFlags() { return kF_Video | kF_Audio; }

	int GetDefaultPriority() { return 0; }

	const wchar_t *GetFilenamePattern() {
#ifdef DISABLE_ASF
		return NULL;
#else
		return L"Windows Media format (*.asf,*.wmv)\0*.asf;*.wmv\0";
#endif
	}

	bool DetectByFilename(const wchar_t *pszFilename) {
		return false;
	}

	int DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize) {
		if (nHeaderSize >= 16) {
			if (!memcmp(pHeader, &GUID_ASF_Header_Object, 16))
				return 1;
		}

		return -1;
	}

	InputFile *CreateInputFile(uint32 flags) {
#ifdef DISABLE_ASF
		throw MyError("Not supported: Microsoft owns US patent #6,041,345 on the ASF file format, preventing "
					"third-party applications from extracting data from ASF files. ASF file format support "
					"was removed as of V1.3d at the request of Microsoft and to avoid patent "
					"infringement claims, and as such VirtualDub no longer supports ASF. Please "
					"do not ask for versions that do.");
#else
		InputFile *pf = new_nothrow InputFileASF;

		if (!pf)
			throw MyMemoryError();

		return pf;
#endif
	}

};

extern IVDInputDriver *VDCreateInputDriverASF() { return new VDInputDriverASF; }
